from .....model import TextGenerationVendor
from .....model.nlp.text.vendor.openai import OpenAIClient, OpenAIModel
from transformers import PreTrainedModel


class LiteLLMClient(OpenAIClient):
    def __init__(self, api_key: str | None = None, base_url: str | None = None):
        super().__init__(
            api_key=api_key or "",
            base_url=base_url or "http://localhost:4000",
        )


class LiteLLMModel(OpenAIModel):
    def _load_model(self) -> PreTrainedModel | TextGenerationVendor:
        return LiteLLMClient(
            base_url=self._settings.base_url,
            api_key=self._settings.access_token,
        )



from ..memory import RecentMessageMemory
from ..memory.partitioner.text import TextPartitioner
from ..memory.permanent import PermanentMessageMemory, VectorFunction
from ..model.entities import EngineMessage
from typing import Any, Optional, Type
from uuid import UUID


class MemoryManager:
    _agent_id: UUID
    _participant_id: UUID
    _permanent_message_memory: Optional[PermanentMessageMemory] = None
    _recent_message_memory: Optional[RecentMessageMemory] = None
    _text_partitioner: TextPartitioner

    @classmethod
    async def create_instance(
        cls,
        *args,
        agent_id: UUID,
        participant_id: UUID,
        text_partitioner: TextPartitioner,
        with_permanent_message_memory: Optional[str] = None,
        with_recent_message_memory: bool = True,
    ):
        permanent_memory: Optional[PermanentMessageMemory] = None
        if with_permanent_message_memory:
            from .permanent.pgsql import PgsqlMessageMemory

            permanent_memory = await PgsqlMessageMemory.create_instance(
                dsn=with_permanent_message_memory
            )
        recent_memory = (
            RecentMessageMemory() if with_recent_message_memory else None
        )

        manager = cls(
            agent_id=agent_id,
            participant_id=participant_id,
            permanent_message_memory=permanent_memory,
            recent_message_memory=recent_memory,
            text_partitioner=text_partitioner,
        )
        return manager

    def __init__(
        self,
        *args,
        agent_id: UUID,
        participant_id: UUID,
        permanent_message_memory: Optional[PermanentMessageMemory],
        recent_message_memory: Optional[RecentMessageMemory],
        text_partitioner: TextPartitioner,
    ):
        assert agent_id and participant_id
        self._agent_id = agent_id
        self._participant_id = participant_id
        self._text_partitioner = text_partitioner
        if permanent_message_memory:
            self.add_permanent_message_memory(permanent_message_memory)
        if recent_message_memory:
            self.add_recent_message_memory(recent_message_memory)

    @property
    def has_permanent_message(self) -> bool:
        return bool(self._permanent_message_memory)

    @property
    def has_recent_message(self) -> bool:
        return bool(self._recent_message_memory)

    @property
    def permanent_message(self) -> Optional[PermanentMessageMemory]:
        return self._permanent_message_memory

    @property
    def recent_message(self) -> Optional[RecentMessageMemory]:
        return self._recent_message_memory

    @property
    def recent_messages(self) -> Optional[list[EngineMessage]]:
        return (
            self._recent_message_memory.data
            if self._recent_message_memory
            else None
        )

    def add_recent_message_memory(self, memory: RecentMessageMemory):
        self._recent_message_memory = memory

    def add_permanent_message_memory(self, memory: PermanentMessageMemory):
        self._permanent_message_memory = memory

    async def append_message(self, engine_message: EngineMessage) -> None:
        assert (
            isinstance(engine_message, EngineMessage)
            and engine_message.agent_id
            and engine_message.message
            and engine_message.message.content
        )

        if self._permanent_message_memory:
            partitions = await self._text_partitioner(
                engine_message.message.content
            )
            await self._permanent_message_memory.append_with_partitions(
                engine_message, partitions=partitions
            )

        if self._recent_message_memory:
            self._recent_message_memory.append(engine_message)

    async def continue_session(
        self,
        session_id: UUID,
        *args,
        load_recent_messages: bool = True,
        load_recent_messages_limit: Optional[int] = None,
    ) -> None:
        if self._permanent_message_memory:
            await self._permanent_message_memory.continue_session(
                agent_id=self._agent_id,
                participant_id=self._participant_id,
                session_id=session_id,
            )

        if (
            load_recent_messages
            and self._permanent_message_memory
            and self._recent_message_memory
        ):
            messages = await self._permanent_message_memory.get_recent_messages(
                participant_id=self._participant_id,
                session_id=session_id,
                limit=load_recent_messages_limit,
            )
            self._recent_message_memory.reset()
            for message in messages:
                self._recent_message_memory.append(message)

    async def start_session(self) -> None:
        if self._permanent_message_memory:
            await self._permanent_message_memory.reset_session(
                agent_id=self._agent_id, participant_id=self._participant_id
            )

        if self._recent_message_memory:
            self._recent_message_memory.reset()

    async def search_messages(
        self,
        search: str,
        agent_id: UUID,
        session_id: UUID,
        participant_id: UUID,
        *args,
        function: VectorFunction,
        limit: Optional[int] = None,
    ) -> list[EngineMessage]:
        assert self._permanent_message_memory
        search_partitions = await self._text_partitioner(search)
        messages = await self._permanent_message_memory.search_messages(
            search_partitions=search_partitions,
            agent_id=agent_id,
            session_id=session_id,
            participant_id=participant_id,
            function=function,
            limit=limit,
        )
        return messages

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[Any],
    ):
        pass



from argparse import Namespace
from asyncio import as_completed, create_task, gather, sleep, to_thread
from ...agent.orchestrator import Orchestrator
from ...cli import get_input, confirm
from ...cli.commands.cache import cache_delete, cache_download
from ...event import EventStats
from ...model import TextGenerationResponse
from ...model.entities import GenerationSettings, Model, Token
from ...model.hubs.huggingface import HuggingfaceHub
from ...model.manager import ModelManager
from ...model.criteria import KeywordStoppingCriteria
from ...model.nlp.sentence import SentenceTransformerModel
from ...model.nlp.text.generation import TextGenerationModel
from ...secrets import KeyringSecrets
from . import get_model_settings
from rich.prompt import Prompt
from logging import Logger
from rich.console import Console, Group, RenderableType
from rich.live import Live
from rich.padding import Padding
from rich.theme import Theme
from time import perf_counter
from typing import Tuple, Union


def model_display(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    logger: Logger,
    *vargs,
    is_sentence_transformer: bool | None = None,
    load: bool | None = None,
    model: Union[SentenceTransformerModel, TextGenerationModel] | None = None,
    summary: bool | None = None,
) -> None:
    assert args.model
    _ = theme._

    with ModelManager(hub, logger) as manager:
        engine_uri = manager.parse_uri(args.model)
        model_id = args.model
        can_access = args.skip_hub_access_check or hub.can_access(model_id)
        hub_model = hub.model(model_id)
        console.print(
            theme.model(
                hub_model,
                can_access=can_access,
                expand=(summary is not None and not summary)
                or (summary is None and not args.summary),
                summary=False,
            )
        )

        if not model and (
            (load is not None and load) or (load is None and args.load)
        ):
            model_settings = get_model_settings(
                args,
                hub,
                logger,
                engine_uri,
                is_sentence_transformer=is_sentence_transformer,
            )
            with manager.load(**model_settings) as lm:
                logger.debug(f"Loaded model {lm.config.__repr__()}")
                console.print(
                    Padding(
                        theme.model_display(
                            lm.config,
                            lm.tokenizer_config,
                            summary=summary or False,
                        ),
                        pad=(0, 0, 0, 0),
                    )
                )
        elif model:
            console.print(
                Padding(
                    theme.model_display(
                        model.config,
                        model.tokenizer_config,
                        summary=summary or False,
                    ),
                    pad=(0, 0, 0, 0),
                )
            )


def model_install(
    args: Namespace, console: Console, theme: Theme, hub: HuggingfaceHub
) -> None:
    assert args.model
    engine_uri = ModelManager.parse_uri(args.model)
    if (
        engine_uri.vendor
        and engine_uri.password
        and engine_uri.user == "secret"
    ):
        secrets = KeyringSecrets()
        token = secrets.read(engine_uri.password)
        if token is None:
            secret_value = Prompt.ask(
                theme.ask_secret_password(engine_uri.password)
            )
            secrets.write(engine_uri.password, secret_value)
        elif confirm(console, theme.ask_override_secret(engine_uri.password)):
            secret_value = Prompt.ask(
                theme.ask_secret_password(engine_uri.password)
            )
            secrets.write(engine_uri.password, secret_value)

    cache_download(args, console, theme, hub)


async def model_run(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    logger: Logger,
) -> None:
    assert args.model and args.device and args.max_new_tokens
    _, _i = theme._, theme.icons

    system_prompt = args.system or None

    with ModelManager(hub, logger) as manager:
        engine_uri = manager.parse_uri(args.model)
        model_settings = get_model_settings(
            args, hub, logger, engine_uri, is_sentence_transformer=False
        )

        if not args.quiet:
            if engine_uri.is_local:
                can_access = (
                    args.quiet
                    or args.skip_hub_access_check
                    or hub.can_access(engine_uri.model_id)
                )

                model = hub.model(engine_uri.model_id)
                console.print(
                    Padding(
                        theme.model(model, can_access=can_access, summary=True),
                        pad=(0, 0, 1, 0),
                    )
                )

        with manager.load(**model_settings) as lm:
            logger.debug(f"Loaded model {lm.config.__repr__()}")

            input_string = get_input(
                console,
                _i["user_input"] + " ",
                echo_stdin=not args.no_repl,
                is_quiet=args.quiet,
            )
            if not input_string:
                return

            settings = GenerationSettings(
                do_sample=args.do_sample,
                enable_gradient_calculation=args.enable_gradient_calculation,
                max_new_tokens=args.max_new_tokens,
                min_p=args.min_p,
                repetition_penalty=args.repetition_penalty,
                temperature=args.temperature,
                top_k=args.top_k,
                top_p=args.top_p,
                use_cache=args.use_cache,
            )

            display_tokens = args.display_tokens or 0
            dtokens_pick = 10 if display_tokens > 0 else 0

            if engine_uri.is_local:
                stopping_criteria = (
                    KeywordStoppingCriteria(args.stop_on_keyword, lm.tokenizer)
                    if args.stop_on_keyword
                    else None
                )
                output_generator = lm(
                    input_string,
                    system_prompt=system_prompt,
                    settings=settings,
                    stopping_criterias=[stopping_criteria]
                    if stopping_criteria
                    else None,
                    manual_sampling=display_tokens,
                    pick=dtokens_pick,
                    skip_special_tokens=args.quiet or args.skip_special_tokens,
                )
            else:
                output_generator = lm(
                    input_string,
                    system_prompt=system_prompt,
                    settings=settings,
                )

            await token_generation(
                args=args,
                console=console,
                theme=theme,
                logger=logger,
                orchestrator=None,
                event_stats=None,
                lm=lm,
                input_string=input_string,
                response=await output_generator,
                dtokens_pick=dtokens_pick,
                display_tokens=display_tokens,
                with_stats=not args.quiet,
            )


async def model_search(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    refresh_per_second: int,
) -> None:
    assert args.limit
    _ = theme._

    models: list[Model] = []
    model_access: dict[str, bool] = {}

    # Fetch matching models
    with console.status(
        _("Loading models..."),
        spinner=theme.get_spinner("downloading"),
        refresh_per_second=refresh_per_second,
    ):
        models = [
            model
            for model in hub.models(
                filter=args.filter or None,
                search=args.search or None,
                limit=args.limit,
            )
        ]

    # Tasks to check model access
    tasks = [
        create_task(to_thread(lambda id=model.id: (id, hub.can_access(id))))
        for model in models
    ]

    def _render(
        models: list[Model], model_access: dict[str, bool]
    ) -> list[RenderableType]:
        return [
            theme.model(
                model,
                can_access=model_access[model.id]
                if model.id in model_access
                else None,
            )
            for model in models
        ]

    # Keep list of models updated as tasks are completed
    with Live(
        Group(*_render(models, model_access)),
        console=console,
        refresh_per_second=refresh_per_second,
    ) as live:
        for completed_task in as_completed(tasks):
            model_id, can_access = await completed_task
            model_access[model_id] = can_access

            live.update(Group(*_render(models, model_access)))

        await gather(*tasks)


def model_uninstall(
    args: Namespace, console: Console, theme: Theme, hub: HuggingfaceHub
) -> None:
    assert args.model
    engine_uri = ModelManager.parse_uri(args.model)
    if (
        engine_uri.vendor
        and engine_uri.password
        and engine_uri.user == "secret"
    ):
        secrets = KeyringSecrets()
        secrets.delete(engine_uri.password)

    cache_delete(args, console, theme, hub, is_full_deletion=True)


async def token_generation(
    args: Namespace,
    console: Console,
    theme: Theme,
    logger: Logger,
    orchestrator: Orchestrator | None,
    event_stats: EventStats | None,
    lm: TextGenerationModel,
    input_string: str,
    response: TextGenerationResponse,
    *,
    display_tokens: int,
    dtokens_pick: int,
    with_stats: bool = True,
):
    # If no statistics needed, return as early as possible
    if not with_stats:
        async for token in response:
            text_token = token.token if isinstance(token, Token) else token
            console.print(text_token, end="")
        return

    # From here on, display includes stats and may include token probabilities

    display_time_to_n_token = args.display_time_to_n_token or 256
    display_pause = (
        args.display_pause
        if args.display_pause and args.display_pause > 0
        else 0
    )
    start_thinking = (
        args.start_thinking if hasattr(args, "start_thinking") else False
    )
    tokens = []
    text_tokens = []
    total_tokens = 0
    frame_minimum_pause_ms = (
        100 if display_pause > 0 and display_tokens > 0 else 0
    )

    with Live() as live:
        start = perf_counter()
        input_token_count = (
            response.input_token_count
            if response.input_token_count
            else orchestrator.input_token_count
            if orchestrator
            else lm.input_token_count(input_string)
        )
        ttft: float | None = None
        ttnt: float | None = None
        token_frame_list: list[Tuple[Token | None, RenderableType]] = None
        last_current_dtoken: Token | None = None

        async for token in response:
            text_token = token.token if isinstance(token, Token) else token

            total_tokens = total_tokens + 1
            ellapsed = perf_counter() - start
            if ttft is None:
                ttft = ellapsed
            if ttnt is None and total_tokens >= display_time_to_n_token:
                ttnt = ellapsed
            text_tokens.append(text_token)

            if display_tokens and isinstance(token, Token):
                tokens.append(token)

            token_frames_promise = theme.tokens(
                lm.model_id,
                lm.tokenizer_config.tokens if lm.tokenizer_config else None,
                lm.tokenizer_config.special_tokens
                if lm.tokenizer_config
                else None,
                display_tokens,
                args.display_probabilities if dtokens_pick > 0 else False,
                dtokens_pick,
                # Which tokens to mark as interesting
                lambda dtoken: (
                    dtoken.probability < args.display_probabilities_maximum
                    or len(
                        [
                            t
                            for t in dtoken.tokens
                            if t.id != dtoken.id
                            and t.probability
                            >= args.display_probabilities_sample_minimum
                        ]
                    )
                    > 0
                )
                if display_tokens
                and args.display_probabilities
                and args.display_probabilities_maximum > 0
                and args.display_probabilities_maximum > 0
                else None,
                text_tokens,
                tokens or None,
                input_token_count,
                total_tokens,
                ttft,
                ttnt,
                ellapsed,
                console.width,
                logger,
                event_stats,
                height=6,
                maximum_frames=1,
                start_thinking=start_thinking,
            )

            token_frame_list = [
                token_frame async for token_frame in token_frames_promise
            ]

            # We prioritize a single selected dtoken at a time, it being
            # the leftmost  selected which is also guaranteed by setting
            # minimum_frames=1 when calling theme.tokens()
            token_frames = [token_frame_list[0]]

            for current_dtoken, frame in token_frames:
                live.update(frame)
                if current_dtoken and current_dtoken != last_current_dtoken:
                    last_current_dtoken = current_dtoken
                    if display_pause > 0:
                        await sleep(display_pause / 1000)
                    elif frame_minimum_pause_ms > 0:
                        await sleep(frame_minimum_pause_ms / 1000)
                elif (
                    dtokens_pick > 0
                    and not args.display_probabilities
                    and display_pause > 0
                ):
                    await sleep(display_pause / 1000)

        if (
            dtokens_pick > 0
            and args.display_probabilities
            and token_frame_list
            and len(token_frame_list) > 0
        ):
            for current_dtoken, frame in token_frame_list[1:]:
                live.update(frame)
                if current_dtoken and display_pause > 0:
                    await sleep(display_pause / 1000)
                elif frame_minimum_pause_ms > 0:
                    await sleep(frame_minimum_pause_ms / 1000)



from packaging.version import parse, Version
from pathlib import Path
from tomllib import load
from urllib.parse import urlparse
from urllib.parse import ParseResult


def _config() -> dict:
    project_path = Path(__file__).resolve().parents[2] / "pyproject.toml"
    with open(project_path, "rb") as file:
        config = load(file)
        return config


config = _config()

def license() -> str:
    assert (
        "project" in config
        and "license" in config["project"]
        and "text" in config["project"]["license"]
    )
    return config["project"]["license"]["text"]


def name() -> str:
    assert "project" in config and "name" in config["project"]
    return config["project"]["name"]


def version() -> Version:
    assert (
        "tool" in config
        and "poetry" in config["tool"]
        and "version" in config["tool"]["poetry"]
    )
    version = config["tool"]["poetry"]["version"]
    return parse(version)


def site() -> ParseResult:
    assert (
        "project" in config
        and "urls" in config["project"]
        and "homepage" in config["project"]["urls"]
    )
    homepage = config["project"]["urls"]["homepage"]
    return urlparse(homepage)



from ...compat import override
from ...model import TextGenerationVendor
from ...model.entities import Input
from ...model.nlp import BaseNLPModel
from contextlib import nullcontext
from numpy import ndarray
from torch import no_grad
from transformers import PreTrainedModel
from transformers.tokenization_utils_base import BatchEncoding
from typing import Literal


class SentenceTransformerModel(BaseNLPModel):
    @property
    def supports_sample_generation(self) -> bool:
        return False

    @property
    def supports_token_streaming(self) -> bool:
        return False

    @property
    def uses_tokenizer(self) -> bool:
        return True

    def token_count(self, input: str) -> int:
        token_ids = self.tokenizer.encode(input, add_special_tokens=False)
        return len(token_ids)

    def _load_model(self) -> PreTrainedModel | TextGenerationVendor:
        from sentence_transformers import SentenceTransformer

        model = SentenceTransformer(
            self._model_id,
            cache_folder=self._settings.cache_dir,
            device=self._device,
            trust_remote_code=self._settings.trust_remote_code,
            local_files_only=self._settings.local_files_only,
            token=self._settings.access_token,
            model_kwargs={
                "attn_implementation": self._settings.attention,
                "torch_dtype": BaseNLPModel._get_weight_type(
                    self._settings.weight_type
                ),
                "low_cpu_mem_usage": True
                if self._device
                else self._settings.low_cpu_mem_usage,
                "device_map": self._device,
            },
            backend="torch",
            similarity_fn_name=None,
            truncate_dim=None,
        )
        return model

    def _tokenize_input(
        self,
        input: Input,
        system_prompt: str | None,
        context: str | None,
        tensor_format: Literal["pt"] = "pt",
    ) -> BatchEncoding:
        raise NotImplementedError()

    @override
    async def __call__(
        self, input: Input, *args, enable_gradient_calculation: bool = False
    ) -> ndarray:
        assert self._model, (
            f"Model {self._model} can't be executed, it "
            + "needs to be loaded first"
        )
        assert isinstance(input, str) or isinstance(input, list)
        with no_grad() if not enable_gradient_calculation else nullcontext():
            embeddings = self._model.encode(input, convert_to_numpy=True)
        return embeddings



from ..model.entities import (
    EngineSettings,
    EngineUri,
    GenerationSettings,
    TransformerEngineSettings,
)
from dataclasses import dataclass, field
from enum import StrEnum
from typing import Optional, Union


class NoOperationAvailableException(Exception):
    pass


class EngineType(StrEnum):
    TEXT_GENERATION = "text_generation"


class InputType(StrEnum):
    TEXT = "text"


class OutputType(StrEnum):
    JSON = "json"
    TEXT = "text"


@dataclass(frozen=True, kw_only=True)
class Goal:
    task: str
    instructions: list[str]


@dataclass(frozen=True, kw_only=True)
class Role:
    persona: list[str]


@dataclass(frozen=True, kw_only=True)
class Specification:
    role: Role
    goal: Optional[Goal]
    rules: list[str] = field(default_factory=list)
    input_type: InputType = InputType.TEXT
    output_type: OutputType = OutputType.TEXT
    settings: Optional[GenerationSettings] = None
    template_id: Optional[str] = None
    template_vars: Optional[dict] = None


@dataclass(frozen=True, kw_only=True)
class EngineEnvironment:
    engine_uri: EngineUri
    settings: Union[EngineSettings, TransformerEngineSettings]
    type: EngineType = EngineType.TEXT_GENERATION


@dataclass(frozen=True, kw_only=True)
class Operation:
    specification: Specification
    environment: EngineEnvironment



from dataclasses import dataclass
from enum import StrEnum
from typing import Any


class EventType(StrEnum):
    START = "start"
    TOOL_PROCESS = "tool_process"
    TOOL_EXECUTE = "tool_execute"
    TOOL_RESULT = "tool_result"
    END = "end"
    STREAM_END = "stream_end"
    CALL_PREPARE_BEFORE = "call_prepare_before"
    CALL_PREPARE_AFTER = "call_prepare_after"
    MEMORY_APPEND_BEFORE = "memory_append_before"
    MEMORY_APPEND_AFTER = "memory_append_after"
    MODEL_EXECUTE_BEFORE = "model_execute_before"
    MODEL_EXECUTE_AFTER = "model_execute_after"
    INPUT_TOKEN_COUNT_BEFORE = "input_token_count_before"
    INPUT_TOKEN_COUNT_AFTER = "input_token_count_after"
    TOKEN_GENERATED = "token_generated"


@dataclass(frozen=True, kw_only=True)
class Event:
    type: EventType
    payload: dict[str, Any] | None = None


class EventStats:
    triggers: dict[EventType, int] = {}
    total_triggers: int = 0



from ....compat import override
from ....model import TextGenerationVendorStream
from ....model.entities import (
    GenerationSettings,
    Input,
    TransformerEngineSettings,
)
from ....model import TextGenerationModel
from ....tool.manager import ToolManager
from asyncio import to_thread
from dataclasses import replace
from logging import Logger
from mlx_lm import load, generate
from typing import AsyncGenerator


class MlxLmStream(TextGenerationVendorStream):
    """Async wrapper around a synchronous token generator."""

    def __init__(self, generator):
        super().__init__(generator)
        self._iterator = generator

    async def __anext__(self) -> str:
        try:
            chunk = await to_thread(next, self._iterator)
        except StopIteration:
            raise StopAsyncIteration
        return chunk


class MlxLmModel(TextGenerationModel):
    """Text generation model using the ``mlx_lm`` backend."""

    def __init__(
        self,
        model_id: str,
        settings: TransformerEngineSettings | None = None,
        logger: Logger | None = None,
    ) -> None:
        settings = settings or TransformerEngineSettings()
        if settings.auto_load_tokenizer:
            settings = replace(settings, auto_load_tokenizer=False)
        super().__init__(model_id, settings, logger)

    @property
    def supports_sample_generation(self) -> bool:
        return False

    def _load_model(self):
        assert load, "mlx-lm is not available"
        model, tokenizer = load(self._model_id)
        self._tokenizer = tokenizer
        self._loaded_tokenizer = True
        return model

    def _build_params(self, settings: GenerationSettings) -> dict:
        return {
            "temperature": settings.temperature,
            "top_p": settings.top_p,
            "top_k": settings.top_k,
            "max_tokens": settings.max_new_tokens,
            "stop": settings.stop_strings,
        }

    async def _stream_generator(
        self,
        prompt: str,
        settings: GenerationSettings,
    ) -> AsyncGenerator[str, None]:
        params = self._build_params(settings)
        iterator = generate(
            self._model, self._tokenizer, prompt, stream=True, **params
        )
        stream = MlxLmStream(iter(iterator))
        async for chunk in stream:
            yield chunk

    def _string_output(
        self,
        prompt: str,
        settings: GenerationSettings,
    ) -> str:
        params = self._build_params(settings)
        return generate(self._model, self._tokenizer, prompt, **params)

    @override
    async def __call__(
        self,
        input: Input,
        system_prompt: str | None = None,
        settings: GenerationSettings | None = None,
        *,
        tool: ToolManager | None = None,
    ) -> TextGenerationVendorStream | str:
        settings = settings or GenerationSettings()
        inputs = super()._tokenize_input(
            input,
            system_prompt,
            context=None,
            tensor_format="pt",
            tool=tool,
        )
        prompt = self._tokenizer.decode(
            inputs["input_ids"][0],
            skip_special_tokens=False,
        )
        generation_settings = replace(settings, do_sample=False)
        if settings.use_async_generator:
            return await self._stream_generator(prompt, generation_settings)
        return self._string_output(prompt, generation_settings)



from abc import ABC, abstractmethod
from ...model.engine import Engine
from ...model.entities import ImageEntity
from PIL import Image
from typing import Literal


class BaseVisionModel(Engine, ABC):
    @abstractmethod
    async def __call__(
        self,
        image_source: str | Image.Image,
        tensor_format: Literal["pt"] = "pt",
    ) -> ImageEntity | list[ImageEntity] | list[str]:
        raise NotImplementedError()

    @staticmethod
    def _get_image(image_source: str | Image.Image) -> Image.Image:
        return (
            image_source
            if isinstance(image_source, Image.Image)
            else Image.open(image_source)
        )



from abc import ABC, abstractmethod
from ...agent.orchestrator import Orchestrator
from ...memory.partitioner.text import TextPartition
from ...model.entities import (
    EngineMessage,
    EngineMessageScored,
    HubCache,
    HubCacheDeletion,
    Model,
    ModelConfig,
    SearchMatch,
    SentenceTransformerModelConfig,
    Similarity,
    Token,
    TokenizerConfig,
    User,
)
from ...memory.permanent import Memory
from dataclasses import fields
from datetime import datetime
from enum import StrEnum
from humanize import intcomma, intword, naturalsize, naturaltime
from logging import Logger
from numpy import ndarray
from rich.console import RenderableType
from typing import Callable, Generator, Literal, Optional, Tuple, Union
from uuid import UUID

Formatter = Union[
    Callable[[datetime], str], Callable[[float], str], Callable[[int], str]
]
Formatters = dict[Literal["datetime", "number", "quantity", "size"], Formatter]
Spinner = Literal["cache_accessing", "connecting", "thinking", "downloading"]
Data = StrEnum(
    "Data",
    {
        **{field.name: field.name for field in fields(Model)},
        **{field.name: field.name for field in fields(User)},
    },
)
DataValue = Union[datetime, float, int, str, None]
Stylers = dict[Data, Callable[[DataValue], str]]


class Theme(ABC):
    _all_spinners: dict[Spinner, str]
    _all_stylers: Stylers
    _all_styles: dict[Data, str]
    _icons: dict[Data, str]
    _: Callable[[str], str]

    @property
    def formatters(self) -> Formatters:
        return {}

    @property
    def icons(self) -> dict[Data, str]:
        return {}

    @property
    def quantity_data(self) -> list[str]:
        return []

    @property
    def spinners(self) -> dict[Spinner, str]:
        return {}

    @property
    def stylers(self) -> Stylers:
        return {}

    @property
    def styles(self) -> dict[Data, str]:
        return {}

    @abstractmethod
    def action(
        self,
        name: str,
        description: str,
        author: str,
        model_id: str,
        library_name: str,
        highlight: bool,
        finished: bool,
    ) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def agent(
        self,
        agent: Orchestrator,
        *args,
        models: list[Model | str],
        cans_access: Optional[bool],
    ) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def ask_access_token(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def ask_delete_paths(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def ask_login_to_hub(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def ask_secret_password(self, key: str) -> str:
        raise NotImplementedError()

    @abstractmethod
    def ask_override_secret(self, key: str) -> str:
        raise NotImplementedError()

    @abstractmethod
    def bye(self) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def cache_delete(
        self, cache_deletion: Optional[HubCacheDeletion], deleted: bool
    ) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def cache_list(
        self,
        cache_dir: str,
        cached_models: list[HubCache],
        display_models: Optional[list[str]] = None,
        show_summary: bool = False,
    ) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def download_access_denied(
        self, model_id: str, model_url: str
    ) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def download_start(self, model_id: str) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def download_progress(self) -> tuple[str | RenderableType]:
        raise NotImplementedError()

    @abstractmethod
    def download_finished(self, model_id: str, path: str) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def logging_in(self, domain: str) -> str:
        raise NotImplementedError()

    @abstractmethod
    def memory_embeddings(
        self,
        input_string: str,
        embeddings: ndarray,
        *args,
        total_tokens: int,
        minv: float,
        maxv: float,
        meanv: float,
        stdv: float,
        normv: float,
        embedding_peek: Optional[int] = 3,
        horizontal: bool = True,
        input_string_peek: int = 40,
        show_stats: bool = True,
        partition: Optional[int] = None,
        total_partitions: Optional[int] = None,
    ) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def memory_embeddings_comparison(
        self, similarities: dict[str, Similarity], most_similar: str
    ) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def memory_embeddings_search(
        self, matches: list[SearchMatch], *args, match_preview_length: int = 300
    ) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def memory_partitions(
        self, partitions: list[TextPartition], *args, display_partitions: int
    ) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def model(
        self,
        model: Model,
        *args,
        can_access: Optional[bool] = None,
        expand: bool = False,
        summary: bool = False,
    ) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def model_display(
        self,
        model_config: Optional[
            Union[ModelConfig, SentenceTransformerModelConfig]
        ],
        tokenizer_config: TokenizerConfig,
    ) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def recent_messages(
        self,
        participant_id: UUID,
        agent: Orchestrator,
        messages: list[EngineMessage],
    ):
        raise NotImplementedError()

    @abstractmethod
    def saved_tokenizer_files(
        directory_path: str, total_files: int
    ) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def search_message_matches(
        self,
        participant_id: UUID,
        agent: Orchestrator,
        messages: list[EngineMessageScored],
    ):
        raise NotImplementedError()

    @abstractmethod
    def memory_search_matches(
        self, participant_id: UUID, namespace: str, memories: list[Memory]
    ) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def tokenizer_config(self, config: TokenizerConfig) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    def tokenizer_tokens(
        self,
        dtokens: list[Token],
        added_tokens: Optional[list[str]],
        special_tokens: Optional[list[str]],
        current_dtoken: Optional[Token] = None,
        dtokens_selected: list[Token] = [],
    ) -> RenderableType:
        raise NotImplementedError()

    @abstractmethod
    async def tokens(
        self,
        model_id: str,
        added_tokens: Optional[list[str]],
        special_tokens: Optional[list[str]],
        display_token_size: Optional[int],
        display_probabilities: bool,
        pick: int,
        focus_on_token_when: Optional[Callable[[Token], bool]],
        text_tokens: list[str],
        tokens: Optional[list[Token]],
        input_token_count: int,
        total_tokens: int,
        ttft: float,
        ttnt: float,
        ellapsed: float,
        console_width: int,
        logger: Logger,
        events: Optional[int] = None,
        tool_calls: Optional[int] = None,
        tool_call_results: Optional[int] = None,
        maximum_frames: Optional[int] = None,
        logits_count: Optional[int] = None,
        think_height: int = 6,
        think_padding: int = 1,
        height: int = 12,
        padding: int = 1,
        wrap_padding: int = 4,
    ) -> Generator[Tuple[Optional[Token], RenderableType], None, None]:
        raise NotImplementedError()

    @abstractmethod
    def welcome(
        self,
        url: str,
        name: str,
        version: str,
        license: str,
        user: Optional[User],
    ) -> RenderableType:
        raise NotImplementedError()

    def __init__(
        self,
        translator: Callable[[str], str],
        translator_plurals: Callable[[str, str, int], str],
        formatters: Formatters = {},
        stylers: Stylers = {},
        styles: dict[Data, str] = {},
        spinners: dict[Spinner, str] = {},
        icons: dict[Data, str] = {},
        quantity_data: list[str] = [],
    ):
        data_keys = list(map(str, Data))

        formatters = {
            **{
                "datetime": lambda v: naturaltime(v),
                "number": lambda v: intcomma(v),
                "quantity": lambda v: intword(v),
                "size": lambda v: naturalsize(v),
            },
            **self.formatters,
        }

        all_icons = {
            **{data: None for data in data_keys},
            **self.icons,
            **icons,
        }

        data_keys.extend(all_icons)
        data_keys.extend(self.styles.keys())

        self._all_spinners = {
            **{spinner: None for spinner in Spinner.__args__},
            **self.spinners,
        }
        self._all_stylers = {
            **{
                data: lambda data, value, prefix=None, icon=True: "".join(
                    [
                        prefix or "",
                        f"{all_icons[data]} "
                        if isinstance(icon, bool)
                        and icon
                        and data in self._icons
                        and all_icons[data]
                        else icon
                        if isinstance(icon, str)
                        else "",
                        f"[{data}]",
                        formatters["quantity"](value)
                        if data in quantity_data
                        else formatters["datetime"](value)
                        if isinstance(value, datetime)
                        else formatters["number"](value)
                        if isinstance(value, int) or isinstance(value, float)
                        else value,
                        f"[/{data}]",
                    ]
                )
                for data in data_keys
            },
            **self.stylers,
        }
        self._all_styles = {**{data: "" for data in data_keys}, **self.styles}
        self._icons = all_icons
        self._ = translator
        self._n = translator_plurals

    def get_styles(self) -> dict[Data, str]:
        return self._all_styles

    def get_spinner(self, spinner_name: str) -> Spinner:
        return self._all_spinners[spinner_name]

    def __call__(self, item: Model | str) -> RenderableType:
        return self.model(item) if isinstance(item, Model) else str(item)

    def _f(
        self,
        data: Data,
        value: DataValue,
        prefix: Optional[str] = None,
        icon: Union[bool, str] = True,
    ) -> str:
        return (
            self._all_stylers[data](data, value, prefix, icon)
            if data in self._all_stylers and self._all_stylers[data]
            else str(value)
        )



from ...agent import EngineEnvironment, Goal, Operation, Specification
from ...agent.orchestrator import Orchestrator
from logging import Logger
from ...event.manager import EventManager
from ...memory.manager import MemoryManager
from ...model.entities import EngineUri, TransformerEngineSettings
from ...model.manager import ModelManager
from ...tool.manager import ToolManager
from typing import Optional
from uuid import UUID


class DefaultOrchestrator(Orchestrator):
    def __init__(
        self,
        engine_uri: EngineUri,
        logger: Logger,
        model_manager: ModelManager,
        memory: MemoryManager,
        tool: ToolManager,
        event_manager: EventManager,
        *,
        name: Optional[str],
        role: str,
        task: str,
        instructions: str,
        rules: Optional[list[str]],
        template_id: Optional[str] = None,
        settings: Optional[TransformerEngineSettings] = None,
        call_options: Optional[dict] = None,
        template_vars: Optional[dict] = None,
        id: Optional[UUID] = None,
    ):
        specification = Specification(
            role=role,
            goal=Goal(task=task, instructions=[instructions])
            if task and instructions
            else None,
            rules=rules,
            template_id=template_id or "agent.md",
            template_vars=template_vars,
        )
        super().__init__(
            logger,
            model_manager,
            memory,
            tool,
            event_manager,
            Operation(
                specification=specification,
                environment=EngineEnvironment(
                    engine_uri=engine_uri, settings=settings
                ),
            ),
            call_options=call_options,
            id=id,
            name=name,
        )



from abc import ABC, abstractmethod
from ..compat import override
from ..model import TextGenerationVendor, TokenizerNotSupportedException
from ..model.engine import Engine
from PIL import Image
from torch import argmax, no_grad
from torchaudio import load
from torchaudio.transforms import Resample
from transformers import (
    AutoProcessor,
    AutoModelForCTC,
    PreTrainedModel,
    PreTrainedTokenizer,
    PreTrainedTokenizerFast,
)
from typing import Literal


class BaseAudioModel(Engine, ABC):
    @abstractmethod
    async def __call__(
        self,
        image_source: str | Image.Image,
        tensor_format: Literal["pt"] = "pt",
    ) -> str:
        raise NotImplementedError()

    def _load_tokenizer(
        self, tokenizer_name_or_path: str | None, use_fast: bool = True
    ) -> PreTrainedTokenizer | PreTrainedTokenizerFast:
        raise TokenizerNotSupportedException()

    def _load_tokenizer_with_tokens(
        self, tokenizer_name_or_path: str | None, use_fast: bool = True
    ) -> PreTrainedTokenizer | PreTrainedTokenizerFast:
        raise TokenizerNotSupportedException()


class SpeechRecognitionModel(BaseAudioModel):
    def _load_model(self) -> PreTrainedModel | TextGenerationVendor:
        self._processor = AutoProcessor.from_pretrained(
            self._model_id,
            trust_remote_code=self._settings.trust_remote_code,
            # default behavior in transformers v4.48
            use_fast=True,
        )
        model = AutoModelForCTC.from_pretrained(
            self._model_id,
            trust_remote_code=self._settings.trust_remote_code,
            pad_token_id=self._processor.tokenizer.pad_token_id,
            ctc_loss_reduction="mean",
            device_map=self._device,
        )
        return model

    @override
    async def __call__(
        self,
        audio_source: str,
        sampling_rate: int,
        tensor_format: Literal["pt"] = "pt",
    ) -> str:
        audio_wave, original_sampling_rate = load(audio_source)
        if original_sampling_rate != sampling_rate:
            resampler = Resample(
                orig_freq=original_sampling_rate, new_freq=sampling_rate
            )
            audio_wave = resampler(audio_wave)
        audio = audio_wave.squeeze().numpy()
        inputs = self._processor(
            audio,
            sampling_rate=sampling_rate,
            return_tensors=tensor_format,
        )
        with no_grad():
            # shape (batch, time_steps, vocab_size)
            logits = self._model(inputs.input_values).logits
        predicted_ids = argmax(logits, dim=-1)
        transcription = self._processor.batch_decode(predicted_ids)[0]
        return transcription



from dataclasses import dataclass
from datetime import datetime
from enum import StrEnum
from torch import dtype, Tensor
from typing import Literal, Optional, Union
from uuid import UUID

AttentionImplementation = Literal[
    "eager", "flash_attention_2", "flex_attention", "sdpa"
]

ProbabilityDistribution = Literal[
    "entmax", "gumbel_softmax", "log_softmax", "sparsemax", "softmax"
]

SimilarityFunction = Literal["cosine", "dot", "euclidean", "manhattan"]

ImageTextGenerationLoaderClass = Literal["gemma3", "qwen2"]

TextGenerationLoaderClass = Literal["auto", "gemma3", "mistral3"]

Vendor = Literal[
    "anthropic",
    "anyscale",
    "deepinfra",
    "deepseek",
    "google",
    "groq",
    "huggingface",
    "hyperbolic",
    "local",
    "openai",
    "openrouter",
    "ollama",
    "litellm",
    "together",
]

WeightType = Literal[
    "auto",
    "bool",
    "bf16",
    "f16",
    "f32",
    "f64",
    "i8",
    "i16",
    "i32",
    "i64",
    "ui8",
]

ToolValue = Union[bool, float, int, str, None]


class MessageRole(StrEnum):
    ASSISTANT = "assistant"
    SYSTEM = "system"
    TOOL = "tool"
    USER = "user"


class ToolFormat(StrEnum):
    JSON = "json"
    REACT = "react"
    BRACKET = "bracket"
    OPENAI = "openai"


class DistanceType(StrEnum):
    COSINE = "cosine"
    DOT = "dot"
    L1 = "l1"
    L2 = "l2"
    PEARSON = "pearson"


@dataclass(frozen=True, kw_only=True)
class EngineSettings:
    auto_load_model: bool = True
    auto_load_tokenizer: bool = True
    cache_dir: Optional[str] = None
    change_transformers_logging_level = True
    device: Optional[str] = None
    disable_loading_progress_bar: bool = True
    enable_eval: bool = True
    trust_remote_code: bool = False
    tokenizer_name_or_path: Optional[str] = None


@dataclass(kw_only=True, frozen=True)
class EngineUri:
    host: Optional[str]
    port: Optional[int]
    user: Optional[str]
    password: Optional[str]
    vendor: Optional[Vendor]
    model_id: Optional[str]
    params: dict[str, Union[str, int, float, bool]]

    @property
    def is_local(self) -> bool:
        return not self.vendor or self.vendor == "local"


@dataclass(frozen=True, kw_only=True)
class GenerationSettings:
    # Generation length ------------------------------------------------------
    # The minimum numbers of tokens to generate, ignoring the number of tokens in the prompt.
    min_new_tokens: Optional[int] = None
    # The maximum numbers of tokens to generate, ignoring the number of tokens in the prompt
    max_new_tokens: Optional[int] = None
    # The minimum length of the sequence to be generated. Corresponds to the length of the input prompt + min_new_tokens. Its effect is overridden by min_new_tokens, if also set.
    min_length: Optional[int] = 0
    # The maximum length the generated tokens can have. Corresponds to the length of the input prompt + max_new_tokens. Its effect is overridden by max_new_tokens, if also set.
    max_length: Optional[int] = 20
    # Controls the stopping condition for beam-based methods, like beam-search. It accepts the following values: True, where the generation stops as soon as there are num_beams complete candidates; False, where an heuristic is applied and the generation stops when is it very unlikely to find better candidates; "never", where the beam search procedure only stops when there cannot be better candidates (canonical beam search algorithm)
    early_stopping: Optional[Union[str, bool]] = False
    # The maximum amount of time you allow the computation to run for in seconds. generation will still finish the current pass after allocated time has been passed.
    max_time: Optional[float] = None
    # A string or a list of strings that should terminate generation if the model outputs them.
    stop_strings: Optional[Union[str, list[str]]] = None

    # Generation strategy ----------------------------------------------------
    # Whether or not to use sampling. Use greedy decoding otherwise
    do_sample: bool = False
    # Number of beams for beam search. 1 means no beam search.
    num_beams: Optional[int] = 1
    # Number of groups to divide num_beams into in order to ensure diversity among different groups of beams
    num_beam_groups: Optional[int] = 1
    # The values balance the model confidence and the degeneration penalty in contrastive search decoding
    penalty_alpha: Optional[float] = None

    # Cache ------------------------------------------------------------------
    # Whether or not the model should use the past last key/values attentions (if applicable to the model) to speed up decoding
    use_cache: bool = True

    # Generation output variables --------------------------------------------
    # The number of independently computed returned sequences for each element in the batch.
    num_return_sequences: Optional[int] = 1
    # Whether or not to return the attentions tensors of all attention layers. See attentions under returned tensors for more details.
    output_attentions: Optional[bool] = False
    # Whether or not to return the hidden states of all layers. See hidden_states under returned tensors for more details.
    output_hidden_states: Optional[bool] = False
    # Whether or not to return the prediction scores. See scores under returned tensors for more details.
    output_scores: Optional[bool] = False
    # Whether or not to return the unprocessed prediction logit scores. See logits under returned tensors for more details.
    output_logits: Optional[bool] = None
    # Whether or not to return a ModelOutput, as opposed to returning exclusively the generated sequence. This flag must be set to True to return the generation cache (when use_cache is True) or optional outputs (see flags starting with output_)
    return_dict_in_generate: Optional[bool] = False

    # Output logits manipulation ---------------------------------------------
    # The value used to module the next token probabilities
    temperature: Optional[float] = 1.0
    # The number of highest probability vocabulary tokens to keep for top-k-filtering
    top_k: Optional[int] = 50
    # If set to float < 1, only the smallest set of most probable tokens with probabilities that add up to top_p or higher are kept for generation
    top_p: Optional[float] = 1.0
    # Minimum token probability, which will be scaled by the probability of the most likely token. It must be a value between 0 and 1. Typical values are in the 0.01-0.2 range, comparably selective as setting top_p in the 0.99-0.8 range (use the opposite of normal top_p values)
    min_p: Optional[float] = None
    # The parameter for repetition penalty. 1.0 means no penalty
    repetition_penalty: Optional[float] = 1.0
    # This value is subtracted from a beam’s score if it generates a token same as any beam from other group at a particular time. Note that diversity_penalty is only effective if group beam search is enabled
    diversity_penalty: Optional[float] = 0.0
    # The id of the token to force as the first generated token after the decoder_start_token_id. Useful for multilingual models like mBART where the first generated token needs to be the target language token.
    forced_bos_token_id: Optional[int] = None
    # The id of the token to force as the last generated token when max_length is reached. Optionally, use a list to set multiple end-of-sequence tokens
    forced_eos_token_id: Optional[Union[int, list[int]]] = None

    # Special token usage in generation --------------------------------------
    # The id of the padding token.
    pad_token_id: Optional[int] = None
    # The id of the beginning-of-sequence token
    bos_token_id: Optional[int] = None
    # The id of the end-of-sequence token. Optionally, use a list to set multiple end-of-sequence tokens
    eos_token_id: Optional[Union[int, list[int]]] = None

    # Assistant generation ---------------------------------------------------
    # The number of tokens to be output as candidate tokens.
    prompt_lookup_num_tokens: Optional[int] = None

    # Inference settings -----------------------------------------------------
    # Gradient calculation (set to true when torch.backwards() is called)
    enable_gradient_calculation: bool = False
    # Use async generator (token streaming)
    use_async_generator: bool = True


@dataclass(frozen=True, kw_only=True)
class HubCacheFile:
    name: str
    path: str
    size_on_disk: int
    last_accessed: datetime
    last_modified: datetime


@dataclass(frozen=True, kw_only=True)
class HubCache:
    model_id: str
    path: str
    size_on_disk: int
    revisions: list[str]
    files: dict[str, list[HubCacheFile]]
    total_files: int
    total_revisions: int


@dataclass(frozen=True, kw_only=True)
class HubCacheDeletion:
    model_id: str
    revisions: list[str]
    deletable_size_on_disk: float
    deletable_blobs: list[str]
    deletable_refs: list[str]
    deletable_repos: list[str]
    deletable_snapshots: list[str]


@dataclass(frozen=True, kw_only=True)
class ImageEntity:
    label: str
    score: Optional[float] = None
    box: Optional[list[float]] = None


@dataclass(frozen=True, kw_only=True)
class Message:
    role: MessageRole
    content: str
    name: Optional[str] = None
    arguments: Optional[dict] = None


@dataclass(frozen=True, kw_only=True)
class EngineMessage:
    agent_id: UUID
    model_id: str
    message: Message

    @property
    def is_from_agent(self) -> bool:
        return self.message.role == MessageRole.ASSISTANT


@dataclass(frozen=True, kw_only=True)
class EngineMessageScored(EngineMessage):
    score: float


@dataclass(frozen=True, kw_only=True)
class Model:
    id: str
    parameters: Optional[int]
    parameter_types: Optional[list[str]]
    inference: Optional[str]
    library_name: Optional[str]
    license: Optional[str]
    pipeline_tag: Optional[str]
    tags: list[str]
    architectures: Optional[list[str]]
    model_type: Optional[str]
    auto_model: Optional[str]
    processor: Optional[str]
    gated: Optional[Literal["auto", "manual", False]]
    private: bool
    disabled: Optional[bool]
    last_downloads: int
    downloads: int
    likes: int
    ranking: Optional[int]
    author: str
    created_at: datetime
    updated_at: datetime


@dataclass(frozen=True, kw_only=True)
class ModelConfig:
    # Model architectures that can be used with the model pretrained weights
    architectures: Optional[list[str]]
    # A dict that maps model specific attribute names to the standardized naming of attributes
    attribute_map: dict[str, str]
    # The id of the beginning-of-stream token
    bos_token_id: Optional[str]
    bos_token: Optional[str]
    # If an encoder-decoder model starts decoding with a different token than bos, the id of that token
    decoder_start_token_id: Optional[int]
    # The id of the end-of-stream token
    eos_token_id: Optional[int]
    eos_token: Optional[str]
    # Name of the task used to fine-tune the model. This can be used when converting from an original (TensorFlow or PyTorch) checkpoint
    finetuning_task: Optional[str]
    # The hidden size of the model
    hidden_size: Optional[int]
    # The hidden sizes of the model (ResNet)
    hidden_sizes: Optional[list[int]]
    # A list of keys to ignore by default when looking at dictionary outputs of the model during inference
    keys_to_ignore_at_inference: list[str]
    # The type of loss that the model should use
    loss_type: Optional[str]
    # Maximum input sequence length
    max_position_embeddings: Optional[int]
    # An identifier for the model type
    model_type: str
    # The number of attention heads used in the multi-head attention layers of the model
    num_attention_heads: Optional[int]
    # The number of blocks in the model
    num_hidden_layers: Optional[int]
    # Number of labels to use in the last layer added to the model, typically for a classification task
    num_labels: Optional[int]
    # Whether or not the model should returns all attentions
    output_attentions: bool
    # Whether or not the model should return all hidden-states
    output_hidden_states: bool
    # The id of the padding token
    pad_token_id: Optional[int]
    pad_token: Optional[str]
    # A specific prompt that should be added at the beginning of each text before calling the model
    prefix: Optional[str]
    # The id of the separation token
    sep_token_id: Optional[int]
    sep_token: Optional[str]
    state_size: int
    # Additional keyword arguments to store for the current task
    task_specific_params: Optional[dict[str, any]]
    # The dtype of the weight. Since the config object is stored in plain text, this attribute contains just the floating type string without the torcha
    torch_dtype: dtype
    # The number of tokens in the vocabulary, which is also the first dimension of the embeddings matrix
    vocab_size: Optional[int]
    # The name of the associated tokenizer class to use (if none is set, will use the tokenizer associated to the model by default)
    tokenizer_class: Optional[str]


@dataclass(frozen=True, kw_only=True)
class SearchMatch:
    query: str
    match: str
    l2_distance: float


@dataclass(frozen=True, kw_only=True)
class SentenceTransformerModelConfig:
    backend: Literal["torch", "onnx", "openvino"]
    similarity_function: Optional[SimilarityFunction]
    truncate_dimension: Optional[int]
    transformer_model_config: ModelConfig


@dataclass(frozen=True, kw_only=True)
class Similarity:
    cosine_distance: float
    inner_product: float
    l1_distance: float
    l2_distance: float
    pearson: float


@dataclass(frozen=True, kw_only=True)
class TokenizerConfig:
    name_or_path: str
    tokens: Optional[list[str]]
    special_tokens: Optional[list[str]]
    # Maximum sequence length the tokenizer supports
    tokenizer_model_max_length: int
    # Wether tokenizer is a Rust-based tokenizer
    fast: bool = False


@dataclass(frozen=True, kw_only=True)
class Token:
    id: Union[Tensor, int]
    token: str
    probability: Optional[float] = None


@dataclass(frozen=True, kw_only=True)
class TokenDetail(Token):
    step: Optional[int] = None
    probability_distribution: Optional[ProbabilityDistribution] = None
    tokens: Optional[list[Token]] = None


@dataclass(frozen=True, kw_only=True)
class ToolCall:
    name: str
    arguments: Optional[dict[str, ToolValue]] = None


@dataclass(frozen=True, kw_only=True)
class ToolCallResult(ToolCall):
    call: ToolCall
    result: Optional[ToolValue] = None


@dataclass(frozen=True, kw_only=True)
class QuantizationSettings:
    load_in_4bit: bool
    bnb_4bit_quant_type: Literal["nf4"]
    bnb_4bit_use_double_quant: bool
    bnb_4bit_compute_dtype: type


@dataclass(frozen=True, kw_only=True)
class TransformerEngineSettings(EngineSettings):
    access_token: Optional[str] = None
    attention: Optional[AttentionImplementation] = None
    base_url: Optional[str] = None
    loader_class: Optional[TextGenerationLoaderClass] = "auto"
    local_files_only: bool = False
    low_cpu_mem_usage: bool = False
    quantization: Optional[QuantizationSettings] = None
    revision: Optional[str] = None
    special_tokens: Optional[list[str]] = None
    state_dict: dict[str, Tensor] = None
    tokens: Optional[list[str]] = None
    weight_type: WeightType = "auto"


@dataclass(frozen=True, kw_only=True)
class User:
    name: str
    full_name: Optional[str] = None
    access_token_name: Optional[str] = None


Input = Union[str, list[str], Message, list[Message]]



from asyncio import sleep
from ....compat import override
from ....model import TextGenerationResponse, TextGenerationVendor
from ....model.entities import (
    GenerationSettings,
    Input,
    Message,
    MessageRole,
    ProbabilityDistribution,
    TextGenerationLoaderClass,
    Token,
    TokenDetail,
    TransformerEngineSettings,
)
from ....model.nlp import BaseNLPModel
from ....tool.manager import ToolManager
from dataclasses import replace
from importlib.util import find_spec
from logging import Logger
from threading import Thread
from torch import log_softmax, softmax, topk, Tensor
from torch.nn.functional import gumbel_softmax
from transformers import (
    AsyncTextIteratorStreamer,
    AutoModelForCausalLM,
    Gemma3ForConditionalGeneration,
    Mistral3ForConditionalGeneration,
    PreTrainedModel,
)
from transformers.generation import StoppingCriteria
from transformers.tokenization_utils_base import BatchEncoding
from typing import AsyncGenerator, Literal


class TextGenerationModel(BaseNLPModel):
    _loaders: dict[TextGenerationLoaderClass, type[PreTrainedModel]] = {
        "auto": AutoModelForCausalLM,
        "gemma3": Gemma3ForConditionalGeneration,
        "mistral3": Mistral3ForConditionalGeneration,
    }

    def __init__(
        self,
        model_id: str,
        settings: TransformerEngineSettings | None = None,
        logger: Logger | None = None,
    ) -> None:
        super().__init__(
            model_id, settings or TransformerEngineSettings(), logger
        )

    @property
    def supports_sample_generation(self) -> bool:
        return True

    @property
    def supports_token_streaming(self) -> bool:
        return True

    def _load_model(self) -> PreTrainedModel | TextGenerationVendor:
        assert self._settings.loader_class in self._loaders, (
            f"Unrecognized loader {self._settings.loader_class}"
        )

        if self._settings.quantization and find_spec("bitsandbytes"):
            from transformers import BitsAndBytesConfig

            quantization = self._settings.quantization
            bnb_config = BitsAndBytesConfig(
                load_in_4bit=quantization.load_in_4bit,
                bnb_4bit_quant_type=quantization.bnb_4bit_quant_type,
                bnb_4bit_use_double_quant=quantization.bnb_4bit_use_double_quant,
                bnb_4bit_compute_dtype=quantization.bnb_4bit_compute_dtype,
            )
        else:
            bnb_config = None

        loader = self._loaders[self._settings.loader_class]
        model = loader.from_pretrained(
            self._model_id,
            cache_dir=self._settings.cache_dir,
            attn_implementation=self._settings.attention,
            trust_remote_code=self._settings.trust_remote_code,
            state_dict=self._settings.state_dict,
            local_files_only=self._settings.local_files_only,
            low_cpu_mem_usage=True
            if self._device
            else self._settings.low_cpu_mem_usage,
            torch_dtype=BaseNLPModel._get_weight_type(
                self._settings.weight_type
            ),
            device_map=self._device,
            token=self._settings.access_token,
            quantization_config=bnb_config,
            revision=self._settings.revision,
        )
        return model

    @override
    async def __call__(
        self,
        input: Input,
        system_prompt: str | None = None,
        settings: GenerationSettings | None = None,
        stopping_criterias: list[StoppingCriteria] | None = None,
        *,
        manual_sampling: bool = False,
        pick: int | None = None,
        skip_special_tokens: bool = False,
        tool: ToolManager | None = None,
    ) -> TextGenerationResponse:
        assert self._tokenizer, (
            f"Model {self._model} can't be executed "
            + "without a tokenizer loaded first"
        )
        assert self._model, (
            f"Model {self._model} can't be executed, it "
            + "needs to be loaded first"
        )

        if not settings:
            settings = GenerationSettings()
        assert settings.temperature is None or (
            settings.temperature > 0 or settings.temperature == 0.0
        ), (
            "Temperature has to be a strictly positive float or zero, "
            + "otherwise your next token scores will be invalid"
        )

        do_sample = (
            settings.do_sample if self.supports_sample_generation else False
        )
        if self.supports_sample_generation and settings.temperature:
            do_sample = True

        assert (not do_sample and not settings.temperature) or (
            do_sample and settings.temperature
        ), "Sample-based generation can only be set with temperature"

        output_fn = (
            self._string_output
            if not settings.use_async_generator
            else self._token_generator
            if manual_sampling
            else self._stream_generator
        )
        generation_settings = replace(
            settings,
            do_sample=do_sample,
            pad_token_id=settings.pad_token_id
            if settings.pad_token_id is not None
            else self._tokenizer.eos_token_id,
        )
        inputs = self._tokenize_input(
            input, system_prompt, context=None, tool=tool
        )
        return TextGenerationResponse(
            output_fn,
            inputs=inputs,
            pick=pick,
            settings=generation_settings,
            stopping_criterias=stopping_criterias,
            skip_special_tokens=skip_special_tokens,
            use_async_generator=settings.use_async_generator,
        )

    async def _stream_generator(
        self,
        inputs: dict[str, Tensor] | Tensor,
        settings: GenerationSettings,
        stopping_criterias: list[StoppingCriteria] | None,
        skip_special_tokens: bool,
        **kwargs,
    ) -> AsyncGenerator[str, None]:
        _l = self._log

        streamer = AsyncTextIteratorStreamer(
            self._tokenizer,
            skip_prompt=True,
            decode_kwargs={"skip_special_tokens": skip_special_tokens},
        )

        _l("Created generator async text token streamer")

        def generate_stream() -> None:
            _l(
                f"Streaming up to {settings.max_new_tokens} tokens "
                f"{'with' if settings.do_sample else 'without'} sample "
                f"and {settings.temperature} temperature"
            )
            self._generate_output(
                inputs,
                settings,
                stopping_criterias,
                streamer=streamer,
            )

        thread = Thread(
            target=generate_stream, name=f"{self._model_id}/generate_stream"
        )
        thread.start()

        _l(f"Generation thread #{thread.ident} ({thread.name}) started")

        async for token in streamer:
            yield token

        thread.join()

        _l(f"Generation thread #{thread.ident} ({thread.name}) finished")

    def _string_output(
        self,
        inputs: dict[str, Tensor] | Tensor,
        settings: GenerationSettings,
        stopping_criterias: list[StoppingCriteria] | None,
        skip_special_tokens: bool,
        **kwargs,
    ) -> str:
        _l = self._log
        input_length = inputs["input_ids"].shape[1]
        outputs = self._generate_output(inputs, settings, stopping_criterias)
        response = self._tokenizer.decode(
            outputs[0][input_length:], skip_special_tokens=False
        )
        return response

    async def _token_generator(
        self,
        inputs: dict[str, Tensor] | Tensor,
        settings: GenerationSettings,
        stopping_criterias: list[StoppingCriteria] | None,
        skip_special_tokens: bool,
        pick: int | None,
        probability_distribution: ProbabilityDistribution = "softmax",
    ) -> AsyncGenerator[Token | TokenDetail, None]:
        assert not settings.temperature or (
            settings.temperature >= 0 and settings.temperature <= 1
        ), "temperature should be [0, 1]"
        assert not pick or pick >= 0

        _l = self._log

        enable_entmax = find_spec("entmax") and probability_distribution in [
            "entmax",
            "sparsemax",
        ]
        if enable_entmax:
            import entmax

        _l(
            f"Generating up to {settings.max_new_tokens} tokens "
            f"{'with' if settings.do_sample else 'without'} sample "
            f"and {settings.temperature} temperature"
        )

        generation_settings = replace(
            settings,
            return_dict_in_generate=True,
            output_scores=True,
        )
        outputs = self._generate_output(
            inputs, generation_settings, stopping_criterias
        )
        sequences = outputs.sequences[0]
        scores = outputs.scores  # list of logits for each generated token
        start = inputs["input_ids"].shape[1]  # where generation began
        generated_sequences = sequences[start:]

        _l(f"Generated {len(generated_sequences)} sequences")

        total_tokens = 0
        for step, token_id in enumerate(generated_sequences):
            _l(f"Got step {step} token {token_id}")

            # logits are the raw-unnormalized scores output by the final
            # linear layer
            tensor = scores[step]  # scores is (batch_size, vocab_size)
            logits = tensor[0]  # first element in batch dimension

            # apply probabilty  distribution over last tensor layer, vocab_size
            logits_probs = (
                log_softmax(logits, dim=-1)
                if probability_distribution == "log_softmax"
                else gumbel_softmax(
                    logits, tau=settings.temperature, hard=False, dim=-1
                )
                if probability_distribution == "gumbel_softmax"
                else entmax.sparsemax(logits, dim=-1)
                if enable_entmax and probability_distribution == "sparsemax"
                else entmax.entmax15(logits, dim=-1)
                if enable_entmax and probability_distribution == "entmax"
                else softmax(logits / settings.temperature, dim=-1)
            )

            tokens: list[Token] | None = None
            if pick > 0:
                picked_logits = topk(logits_probs, pick)
                picked_logits_ids = picked_logits.indices.tolist()
                picked_logits_probs = picked_logits.values.tolist()
                tokens = [
                    Token(
                        id=token_id,
                        token=self._tokenizer.decode(
                            token_id, skip_special_tokens=skip_special_tokens
                        ),
                        probability=picked_logits_probs[i],
                    )
                    for i, token_id in enumerate(picked_logits_ids)
                ]

            raw_token = TokenDetail(
                id=token_id,
                token=self._tokenizer.decode(
                    token_id, skip_special_tokens=skip_special_tokens
                ),
                probability=logits_probs[token_id].item(),
                step=step,
                probability_distribution=probability_distribution,
                tokens=tokens,
            )

            _l(f"Yielding step {step} token detail {raw_token.__repr__()}")

            yield raw_token

            total_tokens = total_tokens + 1

        _l(f"Yielded {total_tokens}")

        await sleep(0)  # and just like that, a generator is an async generator

    def _tokenize_input(
        self,
        input: Input,
        system_prompt: str | None,
        context: str | None,
        tensor_format: Literal["pt"] = "pt",
        chat_template: str | None = None,
        tool: ToolManager | None = None,
    ) -> dict[str, Tensor] | BatchEncoding | Tensor:
        _l = self._log
        messages = self._messages(input, system_prompt, tool)
        template_messages = [
            {"role": message.role, "content": message.content}
            for message in messages
        ]

        if not self._tokenizer.chat_template:
            _l("Model does not support template messages, so staying plain")

            prompt = f"{system_prompt}\n\n" or ""
            use_prefix = not any(
                tm["role"] == MessageRole.USER for tm in template_messages
            )

            for template_message in template_messages:
                if use_prefix:
                    prompt += (
                        "User: "
                        if template_message["role"] == MessageRole.USER
                        else "Assistant: "
                        if template_message["role"] == MessageRole.ASSISTANT
                        else ""
                    )
                prompt += template_message["content"].strip() + "\n"

            inputs = self._tokenizer(
                prompt, add_special_tokens=True, return_tensors=tensor_format
            )
        else:
            _l(f"Got {len(template_messages)} template messages")

            _l(f"Applying chat template to {len(template_messages)} messages")
            inputs = self._tokenizer.apply_chat_template(
                template_messages,
                chat_template=chat_template,
                tools=tool.tools if tool else None,
                add_generation_prompt=True,
                tokenize=True,
                add_special_tokens=True,
                return_dict=True,
                return_tensors=tensor_format,
            )

        _l(f"Translating inputs to {self._model.device}")
        inputs = inputs.to(self._model.device)
        return inputs

    def _messages(
        self,
        input: Input,
        system_prompt: str | None,
        tool: ToolManager | None = None,
    ) -> list[Message]:
        if isinstance(input, str):
            input = Message(role=MessageRole.USER, content=input)
        elif isinstance(input, list):
            for m in input:
                assert isinstance(m, Message)
        elif not isinstance(input, Message):
            raise ValueError(input)

        messages = [input] if not isinstance(input, list) else input

        if system_prompt:
            messages = [
                Message(role=MessageRole.SYSTEM, content=system_prompt)
            ] + messages

        assert isinstance(messages, list)
        return messages



from abc import ABC
from ...model.transformer import TransformerModel
from ...model.entities import GenerationSettings, WeightType
from contextlib import nullcontext
from torch import (
    dtype,
    bool as tbool,
    bfloat16,
    float16,
    float32,
    float64,
    int8,
    int16,
    int32,
    int64,
    no_grad,
    Tensor,
    uint8,
)
from transformers import AsyncTextIteratorStreamer
from transformers.generation import StoppingCriteria
from typing import Literal


class BaseNLPModel(TransformerModel, ABC):
    def _generate_output(
        self,
        inputs: dict[str, Tensor] | Tensor,
        settings: GenerationSettings,
        stopping_criterias: list[StoppingCriteria] | None = None,
        streamer: AsyncTextIteratorStreamer | None = None,
    ):
        eos_token_id = (
            settings.eos_token_id
            if settings.eos_token_id
            else self._tokenizer.eos_token_id
            if not settings.forced_eos_token_id and self._tokenizer
            else None
        )
        generation_kwargs = {
            "bos_token_id": settings.bos_token_id,
            "diversity_penalty": settings.diversity_penalty,
            "do_sample": settings.do_sample,
            "early_stopping": settings.early_stopping,
            "eos_token_id": eos_token_id,
            "forced_bos_token_id": settings.forced_bos_token_id,
            "forced_eos_token_id": settings.forced_eos_token_id,
            "max_length": settings.max_length,
            "max_new_tokens": settings.max_new_tokens,
            "max_time": settings.max_time,
            "min_length": settings.min_length,
            "min_new_tokens": settings.min_new_tokens,
            "min_p": settings.min_p,
            "num_beams": settings.num_beams,
            "num_beam_groups": settings.num_beam_groups,
            "num_return_sequences": settings.num_return_sequences,
            "output_attentions": settings.output_attentions,
            "output_hidden_states": settings.output_hidden_states,
            "output_logits": settings.output_logits,
            "output_scores": settings.output_scores,
            "pad_token_id": settings.pad_token_id
            or self._tokenizer.eos_token_id,
            "penalty_alpha": settings.penalty_alpha,
            "prompt_lookup_num_tokens": settings.prompt_lookup_num_tokens,
            "repetition_penalty": settings.repetition_penalty,
            "return_dict_in_generate": settings.return_dict_in_generate,
            "stop_strings": settings.stop_strings,
            "stopping_criteria": stopping_criterias,
            "streamer": streamer,
            "temperature": settings.temperature,
            "top_k": settings.top_k,
            "top_p": settings.top_p,
            "use_cache": settings.use_cache,
        }

        if settings.forced_bos_token_id or settings.forced_eos_token_id:
            del generation_kwargs["bos_token_id"]
            del generation_kwargs["eos_token_id"]

        with (
            no_grad()
            if not settings.enable_gradient_calculation
            else nullcontext()
        ):
            outputs = (
                self._model.generate(
                    inputs, tokenizer=self._tokenizer, **generation_kwargs
                )
                if isinstance(inputs, Tensor)
                else self._model.generate(
                    **inputs, tokenizer=self._tokenizer, **generation_kwargs
                )
            )
        return outputs

    @staticmethod
    def _get_weight_type(weight_type: WeightType) -> Literal["auto"] | dtype:
        wtype = (
            tbool
            if weight_type == "bool"
            else bfloat16
            if weight_type == "bf16"
            else float16
            if weight_type == "f16"
            else float32
            if weight_type == "f32"
            else float64
            if weight_type == "f64"
            else int8
            if weight_type == "i8"
            else int16
            if weight_type == "i16"
            else int32
            if weight_type == "i32"
            else int64
            if weight_type == "i64"
            else uint8
            if weight_type == "ui8"
            else weight_type
            if weight_type == "auto"
            else "auto"
        )
        assert wtype
        return wtype



from ...compat import override
from ...model import TextGenerationVendor
from ...model.nlp import BaseNLPModel
from torch import argmax, no_grad
from transformers import AutoModelForTokenClassification, PreTrainedModel
from transformers.tokenization_utils_base import BatchEncoding
from typing import Literal


class TokenClassificationModel(BaseNLPModel):
    @property
    def supports_sample_generation(self) -> bool:
        return False

    @property
    def supports_token_streaming(self) -> bool:
        return False

    def _load_model(self) -> PreTrainedModel | TextGenerationVendor:
        model = AutoModelForTokenClassification.from_pretrained(
            self._model_id,
            cache_dir=self._settings.cache_dir,
            attn_implementation=self._settings.attention,
            trust_remote_code=self._settings.trust_remote_code,
            torch_dtype=BaseNLPModel._get_weight_type(
                self._settings.weight_type
            ),
            state_dict=self._settings.state_dict,
            local_files_only=self._settings.local_files_only,
            token=self._settings.access_token,
            device_map=self._device,
        )
        return model

    def _tokenize_input(
        self,
        input: input,
        system_prompt: str | None,
        context: str | None,
        tensor_format: Literal["pt"] = "pt",
    ) -> BatchEncoding:
        assert not system_prompt, (
            "Token classification model "
            + f"{self._model_id} does not support chat "
            + "templates"
        )
        _l = self._log
        _l(f"Tokenizing input {input}")
        inputs = self._tokenizer(input, return_tensors=tensor_format)
        inputs = inputs.to(self._model.device)
        return inputs

    @override
    async def __call__(self, input: str) -> dict[str, str]:
        assert self._tokenizer, (
            f"Model {self._model} can't be executed "
            + "without a tokenizer loaded first"
        )
        assert self._model, (
            f"Model {self._model} can't be executed, it "
            + "needs to be loaded first"
        )
        inputs = self._tokenize_input(input, system_prompt=None, context=None)
        with no_grad():
            outputs = self._model(**inputs)
            # logits shape (1, seq_len, num_labels)
            label_ids = argmax(outputs.logits, dim=2)
            tokens = self._tokenizer.convert_ids_to_tokens(
                inputs["input_ids"][0]
            )
            labels = [
                self._model.config.id2label[label_id.item()]
                for label_id in label_ids[0]
            ]
            tokens_to_labels = dict(zip(tokens, labels))
            return tokens_to_labels



from ...agent.orchestrator import Orchestrator
from ...event import EventStats, EventType
from ...memory.partitioner.text import TextPartition
from ...model.entities import (
    EngineMessage,
    EngineMessageScored,
    HubCache,
    HubCacheDeletion,
    Model,
    ModelConfig,
    SearchMatch,
    SentenceTransformerModelConfig,
    Similarity,
    Token,
    TokenizerConfig,
    User,
)
from ...cli.theme import Data, Spinner, Theme
from ...memory.permanent import Memory
from ...utils import _j, _lf
from humanize import clamp, intcomma, intword, naturalday, naturalsize
from locale import format_string
from logging import Logger
from math import ceil
from numpy import ndarray
from numpy.linalg import norm
from os import linesep
from re import sub
from rich import box
from rich.align import Align
from rich.columns import Columns
from rich.console import Group, RenderableType
from rich.padding import Padding
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    TimeElapsedColumn,
    SpinnerColumn,
)
from rich.rule import Rule
from rich.table import Column, Table
from rich.text import Text
from textwrap import wrap
from typing import Callable, Generator, Optional, Tuple, Union
from uuid import UUID


class FancyTheme(Theme):
    @property
    def icons(self) -> dict[Data, str]:
        return {
            "access_token_name": ":lock:",
            "agent_id": ":robot:",
            "agent_output": ":robot:",
            "avalan": ":heavy_large_circle:",
            "author": ":briefcase:",
            "bye": ":vulcan_salute:",
            "can_access": ":white_check_mark:",
            "checking_access": ":mag:",
            "created_at": ":calendar:",
            "disabled": ":cross_mark:",
            "download": ":floppy_disk:",
            "download_access_denied": ":exclamation_mark:",
            "download_finished": ":heavy_check_mark:",
            "downloads": ":floppy_disk:",
            "gated": ":key:",
            "inference": ":brain:",
            "input_token_count": ":laptop_computer:",
            "library_name": ":books:",
            "license": ":balance_scale:",
            "likes": ":orange_heart:",
            "memory": ":brain:",
            "model_id": ":name_badge:",
            "model_type": ":robot_face:",
            "no_access": ":no_entry_sign:",
            "parameters": ":abacus:",
            "pipeline_tag": ":gear:",
            "private": ":closed_lock_with_key:",
            "ranking": ":trophy:",
            "path_blobs": ":file_folder:",
            "path_refs": ":file_folder:",
            "path_repository": ":file_folder:",
            "path_snapshot": ":file_folder:",
            "session": ":card_index_dividers:",
            "task_id": ":robot:",
            "total_tokens": ":abacus:",
            "tokens_rate": ":high_voltage:",
            "events": ":bookmark_tabs:",
            "tool_calls": ":hammer:",
            "tool_call_results": ":package:",
            "ttft": ":seedling:",
            "ttnt": ":alarm_clock:",
            "updated_at": ":calendar:",
            "user": ":hugging_face:",
            "user_input": ":speaking_head:",
            "tags": ":label:",
        }

    @property
    def styles(self) -> dict[Data, str]:
        return {
            "id": "bold",
            "can_access": "green",
            "checking_access": "bright_black blink",
            "created_at": "magenta",
            "downloads": "bright_black",
            "likes": "bright_black",
            "memory": "magenta",
            "memory_embedding_comparison": "dark_orange3",
            "memory_embedding_comparison_similarity": "dark_orange3",
            "memory_embedding_comparison_similarity_high": "bold dark_olive_green3",
            "memory_embedding_comparison_similarity_middle": "orange_red1",
            "memory_embedding_comparison_similarity_low": "dark_red",
            "model_id": "cyan",
            "no_access": "bold red",
            "parameters": "bold cyan",
            "participant_id": "bold",
            "ranking": "bright_black",
            "session_id": "dark_orange3",
            "score": "dark_orange3",
            "tags": "gray30",
            "updated_at": "magenta",
            "user": "bold cyan",
            "version": "bold",
        }

    @property
    def spinners(self) -> dict[Spinner, str]:
        return {
            "agent_loading": "dots12",
            "cache_accessing": "bouncingBar",
            "connecting": "earth",
            "thinking": "dots",
            "downloading": "earth",
        }

    @property
    def quantity_data(self) -> list[str]:
        return ["likes"]

    def action(
        self,
        name: str,
        description: str,
        author: str,
        model_id: str,
        library_name: str,
        highlight: bool,
        finished: bool,
    ) -> RenderableType:
        _i = self._icons
        description_color = (
            "green" if finished else "white" if highlight else "gray62"
        )
        return Panel.fit(
            Padding(
                Group(
                    *_lf(
                        [
                            f"[{description_color}]"
                            f"{description}[/{description_color}]",
                            _i["author"]
                            + (
                                f" [bright_black]{author}[/bright_black]"
                                + " · "
                                + _i["library_name"]
                                + f" [bright_black]{library_name}[/bright_black]"
                            )
                            if highlight
                            else None,
                        ]
                    )
                )
            ),
            title=_i["task_id"] + f" [cyan]{name}[/cyan]",
            subtitle=_i["model_id"]
            + f" [bright_black]{model_id}[/bright_black]",
            box=box.DOUBLE if highlight else box.SQUARE,
        )

    def agent(
        self,
        agent: Orchestrator,
        *args,
        models: list[Model | str],
        can_access: Optional[bool],
    ) -> RenderableType:
        _, _f, _i = self._, self._f, self._icons
        models_group = Group(
            *_lf(
                [
                    _i["model_id"]
                    + " "
                    + ", ".join(
                        [
                            _("{model_id} ({parameters})").format(
                                model_id=_f("model_id", model.id, icon=False),
                                parameters=_f(
                                    "parameters",
                                    _("{n} params").format(
                                        n=self._parameter_count(
                                            model.parameters
                                        )
                                    ),
                                    icon=False,
                                ),
                            )
                            if isinstance(model, Model)
                            else str(model)
                            for model in models
                        ]
                    ),
                    _f(
                        "memory",
                        _j(
                            ", ",
                            [
                                _("short-term message")
                                if agent.memory.has_recent_message
                                else None,
                                _("long-term message ({driver})").format(
                                    driver=type(
                                        agent.memory.permanent_message
                                    ).__name__
                                )
                                if agent.memory.has_permanent_message
                                else None,
                            ],
                            empty=_("stateless"),
                        ),
                    ),
                    _f(
                        "session",
                        " "
                        + _("session: {session_id}").format(
                            session_id=_f(
                                "session_id",
                                str(agent.memory.permanent_message.session_id),
                            )
                        ),
                    )
                    if agent.memory.has_permanent_message
                    and agent.memory.permanent_message.has_session
                    else None,
                ]
            )
        )
        return Panel(
            models_group,
            title=_f("agent_id", agent.name if agent.name else str(agent.id)),
            box=box.DOUBLE,
        )

    def ask_access_token(self) -> str:
        _ = self._
        return _("Enter your Huggingface access token")

    def ask_delete_paths(self) -> str:
        _ = self._
        return _("Delete selected paths?")

    def ask_login_to_hub(self) -> str:
        _ = self._
        return _("Login to huggingface?")

    def ask_secret_password(self, key: str) -> str:
        _ = self._
        return _("Enter secret for {key}").format(key=key)

    def ask_override_secret(self, key: str) -> str:
        _ = self._
        return _("Secret {key} exists, override?").format(key=key)

    def bye(self) -> RenderableType:
        _, _i = self._, self._icons
        return _i["bye"] + " " + _("bye :)")

    def cache_delete(
        self, cache_deletion: Optional[HubCacheDeletion], deleted=False
    ) -> RenderableType:
        _, _f, _n, _i = self._, self._f, self._n, self._icons
        if not cache_deletion or (
            not cache_deletion.deletable_blobs
            and not cache_deletion.deletable_refs
            and not cache_deletion.deletable_repos
            and not cache_deletion.deletable_snapshots
        ):
            return Text(_("Nothing found for deletion. No action taken."))

        total_revisions = len(cache_deletion.revisions)
        elements: list[RenderableType] = []
        if not deleted:
            elements.append(
                _n(
                    "{disk_space} of disk space will be freed after deleting "
                    "{total_revisions} revision for {model_id}",
                    "{disk_space} of disk space will be freed after deleting "
                    "{total_revisions} revisions for {model_id}",
                    total_revisions,
                ).format(
                    model_id=_f("model_id", cache_deletion.model_id),
                    total_revisions=total_revisions,
                    disk_space=naturalsize(
                        cache_deletion.deletable_size_on_disk
                    ),
                )
            )

            for field_name, title, deletable_paths in [
                (
                    "path_blobs",
                    _("BLOBs paths"),
                    cache_deletion.deletable_blobs,
                ),
                ("path_refs", _("Refs paths"), cache_deletion.deletable_refs),
                (
                    "path_repository",
                    _("Repository paths"),
                    cache_deletion.deletable_repos,
                ),
                (
                    "path_snapshot",
                    _("Snapshot paths"),
                    cache_deletion.deletable_snapshots,
                ),
            ]:
                if deletable_paths:
                    panel = Panel(
                        Group(
                            *[_f(field_name, path) for path in deletable_paths]
                        ),
                        title=title,
                    )
                    elements.append(Padding(panel, pad=(1, 0, 0, 0)))
        else:
            elements.append(
                _n(
                    "{disk_space} of disk space were freed after deleting "
                    "{total_revisions} revision for {model_id}",
                    "{disk_space} of disk space were freed after deleting "
                    "{total_revisions} revisions for {model_id}",
                    total_revisions,
                ).format(
                    model_id=_f("model_id", cache_deletion.model_id),
                    total_revisions=total_revisions,
                    disk_space=naturalsize(
                        cache_deletion.deletable_size_on_disk
                    ),
                )
            )

        return Group(*elements)

    def cache_list(
        self,
        cache_dir: str,
        cached_models: list[HubCache],
        display_models: Optional[list[str]] = None,
        show_summary: bool = False,
    ) -> RenderableType:
        _ = self._

        if display_models and not show_summary:
            tables: list[RenderableType] = []
            for model_cache in [
                m for m in cached_models if m.model_id in display_models
            ]:
                table = Table(
                    Column(header=_("Revision"), justify="left"),
                    Column(header=_("File name"), justify="left"),
                    Column(
                        header=_("Size on disk"),
                        footer=naturalsize(model_cache.size_on_disk),
                        justify="left",
                    ),
                    Column(header=_("Last accessed"), justify="left"),
                    Column(header=_("Last modified"), justify="left"),
                    title=model_cache.model_id,
                    caption=model_cache.path,
                    show_footer=True,
                    show_header=True,
                    show_edge=True,
                    show_lines=True,
                    border_style="gray58",
                    caption_style="gray58",
                    footer_style="bold cyan",
                )

                last_revision: Optional[str] = None
                for revision, files in model_cache.files.items():
                    for file in files:
                        new_revision = (
                            last_revision is None or revision != last_revision
                        )
                        if new_revision:
                            last_revision = revision

                        summarized_revision = revision[:6]
                        table.add_row(
                            f"[cyan]{summarized_revision}[/cyan]"
                            if new_revision
                            else f"[bright_black]{summarized_revision}"
                            "[/bright_black]",
                            file.name,
                            f"[bold cyan]{naturalsize(file.size_on_disk)}"
                            "[/bold cyan]",
                            f"[magenta]{naturalday(file.last_accessed)}"
                            "[/magenta]",
                            f"[magenta]{naturalday(file.last_modified)}"
                            "[/magenta]",
                        )

                tables.append(Padding(table, pad=(1, 0, 1, 0)))
            return Group(*tables)
        else:
            display_models = (
                [m for m in cached_models if m.model_id in display_models]
                if display_models
                else cached_models
            )
            total_cache_size = sum([m.size_on_disk for m in display_models])
            table = Table(
                Column(header=_("Model"), justify="left", no_wrap=True),
                Column(header=_("Revisions"), justify="left"),
                Column(header=_("Total files"), justify="left"),
                Column(
                    header=_("Size on disk"),
                    footer=naturalsize(total_cache_size),
                    justify="left",
                ),
                show_footer=True,
                show_header=True,
                show_edge=True,
                show_lines=True,
                border_style="gray58",
                footer_style="bold cyan",
            )

            for model_cache in display_models:
                summarized_revisions = [r[:6] for r in model_cache.revisions]
                table.add_row(
                    model_cache.model_id,
                    f"{model_cache.total_revisions} ([cyan]"
                    + "[/cyan], [cyan]".join(summarized_revisions)
                    + "[/cyan])",
                    intcomma(model_cache.total_files),
                    "[bold cyan]"
                    + naturalsize(model_cache.size_on_disk)
                    + "[/bold cyan]",
                )
            return table

    def download_access_denied(
        self, model_id: str, model_url: str
    ) -> RenderableType:
        _, _i = self._, self._icons
        return Group(
            *_lf(
                [
                    Padding(
                        " ".join(
                            [
                                "[bold red]"
                                + _i["download_access_denied"]
                                + "[/bold red]",
                                "[red]"
                                + _(
                                    "Access denied while trying to download {model_id}"
                                ).format(model_id=model_id)
                                + "[/red]",
                            ]
                        )
                    ),
                    Padding(
                        _(
                            "Ensure you accepted {model_id} license terms: {model_url}"
                        ).format(model_id=model_id, model_url=model_url),
                        pad=(1, 0, 0, 0),
                    ),
                ]
            )
        )

    def download_start(self, model_id: str) -> RenderableType:
        _, _i = self._, self._icons
        return Group(
            _i["download"]
            + " "
            + _("Downloading model {model_id}:").format(model_id=model_id),
            "",
        )

    def download_progress(self) -> tuple[str | RenderableType]:
        _ = self._
        return (
            SpinnerColumn(),
            "[progress.description]{task.description}"
            "[progress.percentage]{task.percentage:>4.0f}%",
            BarColumn(bar_width=None),
            "[",
            MofNCompleteColumn(),
            "-",
            TimeElapsedColumn(),
            "]",
        )

    def download_finished(self, model_id: str, path: str) -> RenderableType:
        _, _i = self._, self._icons
        return Padding(
            " ".join(
                [
                    "[bold green]" + _i["download_finished"] + "[/bold green]",
                    _("Downloaded model {model_id} to {path}").format(
                        model_id=model_id, path=path
                    ),
                ]
            )
        )

    def logging_in(self, domain: str) -> str:
        _ = self._
        return _("Logging in to {domain}...").format(domain=domain)

    def memory_embeddings(
        self,
        input_string: str,
        embeddings: ndarray,
        *args,
        total_tokens: int,
        minv: float,
        maxv: float,
        meanv: float,
        stdv: float,
        normv: float,
        embedding_peek: Optional[int] = 3,
        horizontal: bool = True,
        input_string_peek: int = 30,
        show_stats: bool = True,
        partition: Optional[int] = None,
        total_partitions: Optional[int] = None,
    ) -> RenderableType:
        _ = self._

        assert (
            total_tokens
            and show_stats
            or (embedding_peek and embeddings.size > 2 * embedding_peek)
        )

        peek_table: Optional[Table] = None
        if embedding_peek and embeddings.size > 2 * embedding_peek:
            input_title = (
                input_string[:input_string_peek] + "…"
                if len(input_string) > input_string_peek
                else input_string
            )
            input_title = sub(r"(\r\n|\r|\n)+", " ", input_title)
            peek_title = (
                _("Part #{partition} of #{partitions}: {text}").format(
                    text=input_title.strip(),
                    partition=partition,
                    partitions=total_partitions,
                )
                if partition and total_partitions
                else input_title
            )
            peek_table = Table(
                title=peek_title,
                show_footer=False,
                show_header=True,
                show_edge=True,
                show_lines=True,
                border_style="gray58",
            )

            if horizontal:
                for i in range(embedding_peek):
                    peek_table.add_column(intcomma(i), justify="center")

                peek_table.add_column(_("..."), justify="center")

                for i in range(
                    embeddings.size - embedding_peek, embeddings.size
                ):
                    peek_table.add_column(intcomma(i), justify="center")

                columns = []
                for i, v in enumerate(embeddings[:embedding_peek]):
                    columns.append(clamp(v, format="{:.4g}"))

                columns.append("")

                for i, v in enumerate(embeddings[-embedding_peek:]):
                    columns.append(clamp(v, format="{:.4g}"))

                peek_table.add_row(*columns)
            else:
                peek_table.add_column(_("position"), justify="right")
                peek_table.add_column(_("value"), justify="left")

                for i, v in enumerate(embeddings[:embedding_peek]):
                    peek_table.add_row(intcomma(i), clamp(v, format="{:.4g}"))

                peek_table.add_row(None, _("..."))

                start_i = embeddings.size - embedding_peek
                for i, v in enumerate(embeddings[-embedding_peek:]):
                    peek_table.add_row(
                        intcomma(start_i + i), clamp(v, format="{:.4g}")
                    )

        stats: Optional[Table] = None
        if show_stats:
            stats = Table(
                show_footer=False,
                show_header=True,
                show_edge=True,
                show_lines=True,
                border_style="gray58",
            )
            stats.add_column(_("Token count"))
            stats.add_column(_("Size"))
            stats.add_column(_("Min"))
            stats.add_column(_("Max"))
            stats.add_column(_("Mean"))
            stats.add_column(_("Std"))
            stats.add_column(_("‖v‖"))
            stats.add_row(
                *(
                    clamp(x, format="{:.4g}")
                    if isinstance(x, float)
                    else intcomma(x)
                    for x in (
                        total_tokens,
                        embeddings.size,
                        minv,
                        maxv,
                        meanv,
                        stdv,
                        normv,
                    )
                )
            )

        assert peek_table or stats

        return Group(
            *_lf(
                [
                    Align(peek_table, align="center") if peek_table else None,
                    Align(stats, align="center") if stats else None,
                ]
            )
        )

    def memory_embeddings_comparison(
        self, similarities: dict[str, Similarity], most_similar: str
    ) -> RenderableType:
        assert similarities and most_similar
        _, _f = self._, self._f
        table = Table(
            _("Comparison string"),
            _("Cosine distance"),
            _("L1 distance (Euclidean)"),
            _("L2 distance (Manhattan)"),
            _("Negative dot product"),
            _("Pearson similarity"),
            show_footer=False,
            show_header=True,
            show_edge=True,
            show_lines=True,
            border_style="gray58",
        )
        for compare_string, similarity in similarities.items():
            is_most = compare_string == most_similar
            field_class = (
                "memory_embedding_comparison_similarity_high"
                if similarity.l2_distance <= 0.80
                else "memory_embedding_comparison_similarity_middle"
                if similarity.l2_distance <= 1.2
                else "memory_embedding_comparison_similarity_low"
            )

            table.add_row(
                _f(
                    field_class,
                    compare_string,
                    icon=":trophy: " if is_most else None,
                ),
                _f(
                    field_class,
                    clamp(similarity.cosine_distance, format="{:.4g}"),
                    icon=False,
                ),
                _f(
                    field_class,
                    clamp(similarity.l1_distance, format="{:.4g}"),
                    icon=False,
                ),
                _f(
                    field_class,
                    clamp(similarity.l2_distance, format="{:.4g}"),
                    icon=False,
                ),
                _f(
                    field_class,
                    clamp(similarity.inner_product, format="{:.4g}"),
                    icon=False,
                ),
                _f(
                    field_class,
                    clamp(similarity.pearson, format="{:.4g}"),
                    icon=False,
                ),
            )
        return Align(table, align="center")

    def memory_embeddings_search(
        self, matches: list[SearchMatch], *args, match_preview_length: int = 300
    ) -> RenderableType:
        assert matches
        _, _f = self._, self._f
        table = Table(
            _("Search string"),
            _("Knowledge match"),
            _("L2 distance (Manhattan)"),
            show_footer=False,
            show_header=True,
            show_edge=True,
            show_lines=True,
            border_style="gray58",
        )
        for i, match in enumerate(matches):
            field_class = (
                "memory_embedding_comparison_similarity_high"
                if match.l2_distance <= 0.80
                else "memory_embedding_comparison_similarity_middle"
                if match.l2_distance <= 1.2
                else "memory_embedding_comparison_similarity_low"
            )
            is_most = i == 0
            table.add_row(
                _f(
                    field_class,
                    match.query,
                    icon=":trophy: " if is_most else None,
                ),
                _f(
                    field_class,
                    match.match
                    if len(match.match) <= match_preview_length
                    else match.match[:match_preview_length] + "...",
                    icon=False,
                ),
                _f(
                    field_class,
                    clamp(match.l2_distance, format="{:.4g}"),
                    icon=False,
                ),
            )
        return Align(table, align="center")

    def memory_partitions(
        self, partitions: list[TextPartition], *args, display_partitions: int
    ) -> RenderableType:
        _ = self._
        total_partitions = len(partitions)
        head_count: int = total_partitions
        tail_count: int = 0

        if total_partitions > 2 and total_partitions > display_partitions:
            head_count = ceil((display_partitions - 1) / 2)
            tail_count = display_partitions - head_count

        elements: list[RenderableType] = []

        if head_count:
            for i, partition in enumerate(partitions[:head_count]):
                elements.append(
                    self.memory_embeddings(
                        partition.data,
                        partition.embeddings,
                        total_tokens=partition.total_tokens,
                        minv=partition.embeddings.min().item(),
                        maxv=partition.embeddings.max().item(),
                        meanv=partition.embeddings.mean().item(),
                        stdv=partition.embeddings.std().item(),
                        normv=norm(partition.embeddings).item(),
                        partition=i + 1,
                        total_partitions=total_partitions,
                    )
                )

        if head_count and tail_count:
            elements.append(
                Align(Padding(_("..."), pad=(0, 0, 1, 0)), align="center")
            )

        if tail_count:
            for i, partition in enumerate(partitions[-tail_count:]):
                elements.append(
                    self.memory_embeddings(
                        partition.data,
                        partition.embeddings,
                        total_tokens=partition.total_tokens,
                        minv=partition.embeddings.min().item(),
                        maxv=partition.embeddings.max().item(),
                        meanv=partition.embeddings.mean().item(),
                        stdv=partition.embeddings.std().item(),
                        normv=norm(partition.embeddings).item(),
                        partition=(total_partitions - tail_count) + i + 1,
                        total_partitions=total_partitions,
                    )
                )

        return Group(*elements)

    def model(
        self,
        model: Model,
        *args,
        can_access: Optional[bool] = None,
        expand: bool = False,
        summary: bool = False,
    ) -> RenderableType:
        assert (not expand and not summary) or (expand ^ summary), (
            "From expand and summary, only one can be set"
        )
        _, _f, _i = self._, self._f, self._icons

        return Panel(
            Group(
                *_lf(
                    [
                        _j(
                            " · ",
                            [
                                _j(
                                    " ",
                                    [
                                        _f(
                                            "checking_access",
                                            _("checking access"),
                                        )
                                        if can_access is None
                                        else _f(
                                            "can_access", _("access granted")
                                        )
                                        if can_access
                                        else _f(
                                            "no_access", _("access denied")
                                        ),
                                        _f("author", model.author),
                                        _f("license", model.license)
                                        if expand and model.license
                                        else None,
                                        _f("gated", _("gated"))
                                        if model.gated
                                        else None,
                                        _f("private", _("private"))
                                        if model.private
                                        else None,
                                        _f("disabled", _("disabled"))
                                        if model.disabled
                                        else None,
                                    ],
                                ),
                                (
                                    _i["created_at"]
                                    + " "
                                    + _j(
                                        ", ",
                                        [
                                            _f(
                                                "created_at",
                                                model.created_at,
                                                _("created: "),
                                                icon=False,
                                            )
                                            if expand
                                            else None,
                                            _f(
                                                "updated_at",
                                                model.updated_at,
                                                _("updated: "),
                                                icon=False,
                                            ),
                                        ],
                                    )
                                )
                                if not summary
                                else None,
                            ],
                        ),
                        _j(
                            " · ",
                            [
                                _f("model_type", model.model_type)
                                + (
                                    " (" + ", ".join(model.architectures) + ")"
                                    if expand and model.architectures
                                    else ""
                                )
                                if model.model_type
                                else None,
                                _f("library_name", model.library_name)
                                if model.library_name
                                else None,
                                _f("inference", model.inference)
                                if expand and model.inference
                                else None,
                                _f("pipeline_tag", model.pipeline_tag)
                                if model.pipeline_tag
                                else None,
                            ],
                        )
                        if not summary
                        else None,
                        Rule(style="gray30") if expand and model.tags else None,
                        _f("tags", " " + ", ".join(model.tags))
                        if expand and model.tags
                        else None,
                    ]
                )
            ),
            # Model ID
            title=_f("model_id", model.id)
            + (
                " "
                + _j(
                    " ",
                    [
                        _f(
                            "parameters",
                            self._parameter_count(model.parameters),
                        ),
                        _f("parameter_types", ", ".join(model.parameter_types))
                        if expand and model.parameter_types
                        else None,
                        _("parameters") if expand else _("params"),
                    ],
                )
            )
            if not summary
            else "",
            # Stats
            subtitle=_j(
                " ",
                [
                    _f("downloads", model.downloads)
                    if model.downloads
                    else None,
                    _f("likes", model.likes) if model.likes else None,
                    _f("ranking", model.ranking) if model.ranking else None,
                ],
            )
            if expand
            else None,
            box=box.SQUARE,
        )

    def model_display(
        self,
        model_config: Optional[
            Union[ModelConfig, SentenceTransformerModelConfig]
        ],
        tokenizer_config: TokenizerConfig,
        *args,
        summary: bool = False,
    ) -> RenderableType:
        _ = self._
        return Group(
            *_lf(
                [
                    Padding(
                        self._sentence_transformer_model_config(
                            model_config, summary=summary
                        )
                        if isinstance(
                            model_config, SentenceTransformerModelConfig
                        )
                        else self._model_config(model_config, summary=summary),
                        pad=(0, 0, 0, 0),
                    )
                    if model_config
                    else None,
                    Padding(
                        Panel(
                            self.tokenizer_config(
                                tokenizer_config, summary=summary
                            ),
                            title=_("Tokenizer settings"),
                            border_style="bright_black",
                        ),
                        pad=(0, 0, 0, 0),
                    ),
                ]
            )
        )

    def _sentence_transformer_model_config(
        self, config: SentenceTransformerModelConfig, *args, summary: bool
    ) -> RenderableType:
        _ = self._
        config_table = Table(
            Column(header="", justify="right"),
            Column(header="", justify="left"),
            show_footer=False,
            show_header=False,
            show_edge=True,
            show_lines=True,
            border_style="gray58",
        )
        config_table = self._fill_model_config_table(
            config.transformer_model_config, config_table, summary=summary
        )
        config_table.add_row(_("Backend"), config.backend)
        config_table.add_row(
            _("Similarity function"),
            f"[bold]{config.similarity_function}[/bold]",
        )
        if not summary:
            config_table.add_row(
                _("Truncate dimension"),
                intcomma(config.truncate_dimension)
                if config.truncate_dimension
                else _("No truncation"),
            )

        return Align(config_table, align="center")

    def _model_config(
        self, config: ModelConfig, *args, summary: bool
    ) -> RenderableType:
        config_table = Table(
            Column(header="", justify="right"),
            Column(header="", justify="left"),
            show_footer=False,
            show_header=False,
            show_edge=True,
            show_lines=True,
            border_style="gray58",
        )
        config_table = self._fill_model_config_table(
            config, config_table, summary=summary
        )
        return Align(config_table, align="center")

    def _fill_model_config_table(
        self, config: ModelConfig, config_table: Table, *args, summary: bool
    ) -> Table:
        _ = self._
        config_table.add_row(
            _("Model type"), f"[bold]{config.model_type}[/bold]"
        )
        if not summary and config.architectures:
            config_table.add_row(
                _("Architectures"),
                ", ".join([f"[bold]{a}[/bold]" for a in config.architectures]),
            )

        if config.max_position_embeddings:
            config_table.add_row(
                _("Maximum input sequence length"),
                "[bold magenta]"
                + intcomma(config.max_position_embeddings)
                + "[/bold magenta]",
            )
        if not summary:
            config_table.add_row(
                _("Vocabulary size"),
                f"[magenta]{intcomma(config.vocab_size)}[/magenta]",
            )
        config_table.add_row(
            _("Hidden size"),
            f"[magenta]{intcomma(config.hidden_size)}[/magenta]",
        )
        if not summary:
            config_table.add_row(
                _("Number of hidden layers"),
                f"[magenta]{intcomma(config.num_hidden_layers)}[/magenta]",
            )
            config_table.add_row(
                _("Number of attention heads"),
                f"[magenta]{intcomma(config.num_attention_heads)}[/magenta]",
            )
            config_table.add_row(
                _("Number of labels in last layer"),
                f"[magenta]{intcomma(config.num_labels)}[/magenta]",
            )
            if config.loss_type:
                config_table.add_row(
                    _("Type of loss utilized"), config.loss_type
                )

            config_table.add_row(
                _("Returns all attentions"),
                "[bold]" + _("Yes") + "[/bold]"
                if config.output_attentions
                else _("No"),
            )
            config_table.add_row(
                _("Returns all hidden states"),
                "[bold]" + _("Yes") + "[/bold]"
                if config.output_hidden_states
                else _("No"),
            )

            if config.torch_dtype:
                config_table.add_row(
                    _("Weight data type"),
                    f"[green]{config.torch_dtype}[/green]",
                )
            if config.bos_token_id and config.bos_token:
                config_table.add_row(
                    _("Start of stream token"),
                    "[gray50]#"
                    + str(config.bos_token_id)
                    + f"[/gray50] [cyan]{config.bos_token}[/cyan]",
                )
            if config.eos_token_id and config.eos_token:
                config_table.add_row(
                    _("End of stream token"),
                    "[gray50]#"
                    + str(config.eos_token_id)
                    + f"[/gray50] [cyan]{config.eos_token}[/cyan]",
                )
            if config.sep_token_id and config.sep_token:
                config_table.add_row(
                    _("Separation token"),
                    "[gray50]#"
                    + str(config.sep_token_id)
                    + f"[/gray50] [cyan]{config.sep_token}[/cyan]",
                )
            if config.pad_token_id and config.pad_token:
                config_table.add_row(
                    _("Padding token"),
                    "[gray50]#"
                    + str(config.pad_token_id)
                    + f"[/gray50] [cyan]{config.pad_token}[/cyan]",
                )

            if config.prefix:
                config_table.add_row(
                    _("Mandatory beginning prompt"), config.prefix
                )
        return config_table

    def recent_messages(
        self,
        participant_id: UUID,
        agent: Orchestrator,
        messages: list[EngineMessage],
    ):
        _, _f, _i = self._, self._f, self._icons
        group = Group(
            *_lf(
                [
                    Panel(
                        engine_message.message.content,
                        title=(
                            _i["agent_output"] + " " + _f("id", agent.name)
                            if engine_message.is_from_agent
                            else _i["user_input"]
                            + "  "
                            + _f("participant_id", participant_id)
                        ),
                        title_align="left",
                        expand=True,
                        box=box.SQUARE,
                    )
                    for engine_message in messages
                ]
            )
        )
        return group

    def saved_tokenizer_files(
        self, directory_path: str, total_files: int
    ) -> RenderableType:
        _n = self._n
        return Padding(
            _n(
                "Saved {total_files} tokenizer file to {directory_path}",
                "Saved {total_files} tokenizer files to {directory_path}",
                total_files,
            ).format(total_files=total_files, directory_path=directory_path),
            pad=(1, 0, 0, 0),
        )

    def search_message_matches(
        self,
        participant_id: UUID,
        agent: Orchestrator,
        messages: list[EngineMessageScored],
    ):
        _, _f, _i = self._, self._f, self._icons
        group = Group(
            *_lf(
                [
                    Panel(
                        engine_message.message.content,
                        title=(
                            _i["agent_output"] + " " + _f("id", agent.name)
                            if engine_message.is_from_agent
                            else _i["user_input"]
                            + "  "
                            + _f("participant_id", participant_id)
                        ),
                        title_align="left",
                        subtitle=_("Matching score: {score}").format(
                            score=_f(
                                "score",
                                clamp(engine_message.score, format="{:.8g}"),
                            )
                        ),
                        subtitle_align="left",
                        expand=True,
                        box=box.SQUARE,
                    )
                    for engine_message in messages
                ]
            )
        )
        return group

    def memory_search_matches(
        self, participant_id: UUID, namespace: str, memories: list[Memory]
    ) -> RenderableType:
        _, _f, _i = self._, self._f, self._icons
        group = Group(
            *_lf(
                [
                    Panel(
                        memory.data,
                        title=(
                            _i["memory"] + " " + _f("id", memory.identifier)
                        ),
                        title_align="left",
                        subtitle=_(
                            "Participant: {participant} – Namespace: {ns}"
                        ).format(
                            participant=_f("participant_id", participant_id),
                            ns=_f("id", namespace),
                        ),
                        subtitle_align="left",
                        expand=True,
                        box=box.SQUARE,
                    )
                    for memory in memories
                ]
            )
        )
        return group

    def tokenizer_config(
        self, config: TokenizerConfig, *args, summary: bool = False
    ) -> RenderableType:
        _ = self._

        config_table = Table(
            Column(header="", justify="right"),
            Column(header="", justify="left"),
            show_footer=False,
            show_header=False,
            show_edge=True,
            show_lines=True,
            border_style="gray58",
        )

        config_table.add_row(
            _("Name or path"), f"[bold]{config.name_or_path}[/bold]"
        )
        if not summary:
            if config.tokens:
                config_table.add_row(
                    _("Added tokens"),
                    ", ".join([f"[cyan]{t}[/cyan]" for t in config.tokens]),
                )
            config_table.add_row(
                _("Special tokens"),
                ", ".join([f"[cyan]{t}[/cyan]" for t in config.special_tokens]),
            )
        config_table.add_row(
            _("Maximum sequence length"),
            f"[cyan]{config.tokenizer_model_max_length}[/cyan]",
        )
        config_table.add_row(
            _("Fast (rust based)"),
            "[bold]" + _("Yes") + "[/bold]" if config.fast else _("No"),
        )

        return Align(config_table, align="center")

    def tokenizer_tokens(
        self,
        dtokens: list[Token],
        added_tokens: Optional[list[str]],
        special_tokens: Optional[list[str]],
        display_details: bool = False,
        current_dtoken: Optional[Token] = None,
        dtokens_selected: list[Token] = [],
    ) -> RenderableType:
        # Build token panels
        compact_dtokens = True  # For future configurability
        token_panels = [
            Panel(
                Padding(
                    f"[gray50]#{dtoken.id}[/gray50] {dtoken.token}"
                    if display_details
                    else Text(
                        f"{dtoken.token}",
                        style="white on dark_green"
                        if current_dtoken and dtoken == current_dtoken
                        else None,
                    ),
                    pad=(0, 1, 0, 1),
                ),
                padding=(1 if not compact_dtokens else 0, 0),
                border_style="green on dark_green"
                if current_dtoken and dtoken == current_dtoken
                else "cyan"
                if dtoken in dtokens_selected
                else "magenta"
                if (
                    (added_tokens and dtoken.token in added_tokens)
                    or (special_tokens and dtoken.token in special_tokens)
                )
                else "gray30",
                box=box.SQUARE,
            )
            for dtoken in dtokens
        ]

        # Distribute token panels as columns
        columns = Columns(token_panels, equal=False, expand=False, padding=0)
        return Panel(
            Align(columns, align="center"),
            box=box.MINIMAL,
        )

    async def tokens(
        self,
        model_id: str,
        added_tokens: Optional[list[str]],
        special_tokens: Optional[list[str]],
        display_token_size: Optional[int],
        display_probabilities: bool,
        pick: int,
        focus_on_token_when: Optional[Callable[[Token], bool]],
        text_tokens: list[str],
        tokens: Optional[list[Token]],
        input_token_count: int,
        total_tokens: int,
        ttft: float,
        ttnt: float,
        ellapsed: float,
        console_width: int,
        logger: Logger,
        event_stats: Optional[EventStats] = None,
        maximum_frames: Optional[int] = None,
        logits_count: Optional[int] = None,
        think_height: int = 6,
        think_padding: int = 1,
        height: int = 12,
        padding: int = 1,
        wrap_padding: int = 4,
        limit_think_height: bool = True,
        limit_answer_height: bool = False,
        start_thinking: bool = False,
    ) -> Generator[Tuple[Optional[Token], RenderableType], None, None]:
        _, _f, _l = self._, self._f, logger.debug

        # Prepare data for rendering
        output = "".join(text_tokens)
        pick_first = ceil(pick / 2) if pick > 1 else pick
        max_width = console_width - wrap_padding
        think_wrapped, wrapped = [], []
        is_thinking = start_thinking

        for line in output.splitlines():
            if line == "<think>":
                is_thinking = True
                continue
            elif is_thinking and line == "</think>":
                is_thinking = False
                continue

            wrapped_line = wrap(line, width=max_width)
            if is_thinking:
                think_wrapped.extend(wrapped_line)
                think_wrapped.append(linesep)
            else:
                wrapped.extend(wrapped_line)
                wrapped.append(linesep)

        think_section = (
            think_wrapped[-(think_height - 2 * think_padding) :]
            if think_wrapped and limit_think_height
            else think_wrapped
        )
        think_wrapped_output = (
            " ".join(think_section).strip() if think_section else None
        )

        wrapped_section = (
            wrapped[-(height - padding) :]
            if wrapped and limit_answer_height
            else wrapped
        )
        wrapped_output = (
            " ".join(wrapped_section).strip() if wrapped_section else None
        )

        dtokens = (
            tokens[-display_token_size:]
            if display_token_size and tokens
            else None
        )
        dtokens_selected = (
            [
                dtoken
                for dtoken in dtokens
                if focus_on_token_when and focus_on_token_when(dtoken)
            ]
            if dtokens
            else None
        )
        dtokens_total_selected = (
            len(dtokens_selected) if dtokens_selected else 0
        )

        # Build think and answer panels
        progress_title = " · ".join(
            _lf(
                [
                    _f(
                        "input_token_count",
                        _("{total_tokens} input tokens").format(
                            total_tokens=input_token_count
                        ),
                    ),
                    _f(
                        "total_tokens",
                        _("{total_tokens} output tokens").format(
                            total_tokens=total_tokens
                        ),
                    ),
                    _f("ttft", _("ttft: {ttft} s").format(ttft=f"{ttft:.2f}"))
                    if ttft
                    else None,
                    _f("ttnt", _("ttnt: {ttnt} s").format(ttnt=f"{ttnt:.2f}"))
                    if ttnt
                    else None,
                    _f(
                        "tokens_rate",
                        _("{tokens_rate} t/s").format(
                            tokens_rate=f"{total_tokens / ellapsed:.2f}"
                        ),
                    ),
                    _f(
                        "events",
                        _("{total} events").format(
                            total=event_stats.total_triggers
                        ),
                    )
                    if event_stats
                    else None,
                    _f(
                        "tool_calls",
                        _("{total} tool calls").format(
                            total=event_stats.triggers[EventType.TOOL_EXECUTE]
                        ),
                    )
                    if event_stats
                    and EventType.TOOL_EXECUTE in event_stats.triggers
                    and event_stats.triggers[EventType.TOOL_EXECUTE]
                    else None,
                    _f(
                        "tool_call_results",
                        _("{total} results").format(
                            total=event_stats.triggers[EventType.TOOL_RESULT]
                        ),
                    )
                    if event_stats
                    and EventType.TOOL_RESULT in event_stats.triggers
                    and event_stats.triggers[EventType.TOOL_RESULT]
                    else None,
                ]
            )
        )
        think_pannel = (
            Panel(
                f"[bright_black]{think_wrapped_output}[/bright_black]",
                title=_("{model_id} reasoning").format(
                    model_id=_f("id", model_id)
                ),
                title_align="left",
                subtitle=progress_title if not wrapped_output else None,
                subtitle_align="right",
                height=think_height + 2 * think_padding
                if limit_think_height
                else None,
                padding=think_padding,
                expand=True,
                box=box.SQUARE,
            )
            if think_wrapped_output
            else None
        )
        answer_panel = (
            Panel(
                wrapped_output,
                title=_("{model_id} response").format(
                    model_id=_f("id", model_id)
                )
                if think_wrapped_output is None
                else None,
                title_align="left",
                subtitle=progress_title
                if wrapped_output or not think_wrapped_output
                else None,
                subtitle_align="right",
                height=height + 2 * padding if limit_answer_height else None,
                padding=padding,
                expand=True,
                box=box.SQUARE,
            )
            if wrapped_output
            else None
        )

        # Quick return of no need for token details
        if display_token_size is None or tokens is None:
            yield (
                None,
                Group(
                    *_lf(
                        [
                            think_pannel or None,
                            answer_panel or None,
                        ]
                    )
                ),
            )
            return

        # Deal with token details
        dtokens_selected_last_index_yielded: Optional[int] = None
        yielded_frames = 0
        yield_next_frame = True

        # As long as we need more frames, we keep yielding
        while yield_next_frame:
            current_selected_index: Optional[int] = None
            tokens_distribution_panel: Optional[Panel] = None

            if display_token_size and tokens:
                # Pick current token to highlight
                current_data = None
                current_dtoken: Optional[Token] = None
                if display_probabilities and dtokens_selected:
                    current_selected_index = (
                        0
                        if dtokens_selected_last_index_yielded is None
                        else dtokens_selected_last_index_yielded + 1
                        if dtokens_selected_last_index_yielded + 1
                        < dtokens_total_selected
                        else None
                    )
                    _l(
                        f"Selected {current_selected_index} selected token for "
                        f"yielding, with {dtokens_selected_last_index_yielded} "
                        "being the previous yielded"
                    )

                    current_dtoken = (
                        dtokens_selected[current_selected_index]
                        if current_selected_index is not None
                        else None
                    )
                    current_data = (
                        [t.probability for t in current_dtoken.tokens]
                        if current_dtoken and current_dtoken.tokens
                        else None
                    )
                    if current_dtoken:
                        _l(
                            f'Selected "{current_dtoken.token}" as '
                            "interesting token, with "
                            + clamp(current_dtoken.probability, format="{:.4g}")
                            + f"and {current_dtoken.tokens}"
                        )

                tokens_panel = self.tokenizer_tokens(
                    dtokens,
                    added_tokens,
                    special_tokens,
                    display_details=False,
                    current_dtoken=current_dtoken,
                    dtokens_selected=dtokens_selected,
                )

                # Build bar chart with token alternative probabilities
                chart = None
                if display_probabilities:
                    current_symmetric_indices = (
                        FancyTheme._symmetric_indices(current_data)
                        if current_data
                        else None
                    )
                    current_symmetric_data = (
                        [current_data[i] for i in current_symmetric_indices]
                        if current_data
                        else None
                    )
                    labels = (
                        " ".join(
                            [
                                f"{i + 1}"
                                for i in range(len(current_symmetric_data))
                            ]
                        )
                        if current_symmetric_data
                        else " "
                    )
                    chart_height = 5

                    chart_rows: list[str] = []
                    for level in range(chart_height, 0, -1):
                        chart_row = ""
                        for value in current_symmetric_data or [
                            0 for i in range(pick)
                        ]:
                            if value * chart_height >= level:
                                chart_row += "".join(
                                    ["[green]", "█ ", "[/green]"]
                                )
                            else:
                                chart_row += "  "
                        chart_rows.append(chart_row)
                    chart_rows.append(f"[gray30]{labels}[/gray30]")

                    chart = Align(
                        Panel(
                            Group(*[row for row in chart_rows]),
                            border_style="gray30",
                        ),
                        align="center",
                    )

                # Split token alternatives to two tables
                dbatch_first_table = None
                dbatch_second_table = None
                if pick > 0 and current_dtoken and current_dtoken.tokens:
                    dtoken_tokens = current_dtoken.tokens
                    max_dtoken = max(
                        dtoken_tokens, key=lambda dtoken: dtoken.probability
                    )

                    if pick_first is None or len(dtoken_tokens) <= pick_first:
                        dbatch_first = dtoken_tokens
                        dbatch_second = None
                    else:
                        dbatch_first = dtoken_tokens[:pick_first]
                        dbatch_second = dtoken_tokens[pick_first:]

                    if dbatch_first:
                        dbatch_first_table = self._tokens_table(
                            dbatch_first, current_dtoken, max_dtoken
                        )

                    if dbatch_second:
                        dbatch_second_table = self._tokens_table(
                            dbatch_second, current_dtoken, max_dtoken
                        )

                # Build token distribution panel
                tokens_distribution_panel = Panel(
                    Group(
                        *_lf(
                            [
                                tokens_panel,
                                Align(
                                    Panel.fit(
                                        f" [gray50]#{current_dtoken.id}[/gray50]"
                                        f" {current_dtoken.token} ",
                                        border_style="green",
                                        padding=0,
                                    ),
                                    align="center",
                                )
                                if current_dtoken
                                else None,
                                chart if current_dtoken and chart else None,
                                Align(
                                    Columns(
                                        _lf(
                                            [
                                                dbatch_first_table,
                                                dbatch_second_table,
                                            ]
                                        )
                                    ),
                                    align="center",
                                )
                                if dbatch_first_table or dbatch_second_table
                                else None,
                            ]
                        )
                    )
                    if chart
                    else tokens_panel,
                    title=_("token distribution"),
                    title_align="left",
                    border_style="bright_black",
                )

            yield (
                current_dtoken,
                Group(
                    *_lf(
                        [
                            think_pannel or None,
                            answer_panel or None,
                            tokens_distribution_panel
                            if tokens and tokens_distribution_panel
                            else None,
                        ]
                    )
                ),
            )

            yielded_frames = yielded_frames + 1

            yield_next_frame = (
                not maximum_frames or yielded_frames < maximum_frames
            ) and current_selected_index is not None
            if yield_next_frame:
                dtokens_selected_last_index_yielded = current_selected_index
                _l("Will continue to yield next frame")

    def _tokens_table(
        self,
        dbatch: list[Token],
        current_dtoken: Optional[Token],
        max_dtoken: Optional[Token],
    ):
        _p = self._percentage

        dtable_color = "gray58"
        table = Table(
            show_footer=False,
            show_header=False,
            show_edge=True,
            show_lines=True,
            border_style=dtable_color,
        )
        table.add_column()
        table.add_column(justify="right")
        table.add_column()
        for dtoken in dbatch:
            is_current_token = current_dtoken and current_dtoken.id == dtoken.id
            is_max_dtoken = max_dtoken and max_dtoken.id == dtoken.id
            dtoken_color = (
                "bold green"
                if is_current_token and is_max_dtoken
                else "green"
                if is_current_token
                else "cyan"
                if is_max_dtoken
                else dtable_color
            )
            table.add_row(
                f"[gray50]#{dtoken.id}[/gray50]",
                f"[{dtoken_color}]{dtoken.token}[/{dtoken_color}]",
                f"[{dtoken_color}]{_p(dtoken.probability)}[/{dtoken_color}]",
            )
        return table

    def welcome(
        self,
        url: str,
        name: str,
        version: str,
        license: str,
        user: Optional[User],
    ) -> RenderableType:
        _, _f, _i = self._, self._f, self._icons
        license_text = _("{license} license").format(license=license)
        return Padding(
            Panel(
                Padding(
                    _j(
                        " - ",
                        [
                            " ".join(
                                [
                                    _i["avalan"]
                                    + f" [link={url}]{name}[/link]",
                                    f"[version]{version}[/version]",
                                    "[bright_black]"
                                    + _i["license"]
                                    + f" {license_text}[/bright_black]",
                                ]
                            ),
                            _f("user", user.name) if user else None,
                            _f("access_token_name", user.access_token_name)
                            if user
                            else None,
                        ],
                    )
                ),
                box=box.SQUARE,
            ),
            pad=(0, 0, 0, 0),  # Might bring lower padding back (0,0,1,0)
        )

    def _parameter_count(self, parameters: Optional[int]) -> str:
        _ = self._
        if not parameters:
            return _("N/A")
        return (
            "{:.1f}B".format(parameters / 1e9)
            if parameters >= 1e9
            else intword(parameters)
        )

    @staticmethod
    def _symmetric_indices(data: list[float]) -> list[float]:
        """
        Sorts the data in descending order and arranges them so that
        the highest values are in the center, with lower values at the edges.
        """
        assert data
        sorted_data = sorted(data, reverse=True)
        n = len(sorted_data)
        result = [None] * n

        left = n // 2 - 1
        right = n // 2

        for i, value in enumerate(sorted_data):
            if i % 2 == 0:
                result[left] = i
                left -= 1
            else:
                result[right] = i
                right += 1
        return result

    @staticmethod
    def _percentage(value: float) -> str:
        p = value * 100
        return (
            format_string("%d%%", p, grouping=True)
            if p == int(p)
            else format_string("%.1f%%", p, grouping=True)
        )



from dataclasses import dataclass
from types import FunctionType
from typing import Sequence


@dataclass(frozen=True, kw_only=True)
class ToolSet:
    """Collection of tools sharing an optional namespace."""

    tools: Sequence[FunctionType]
    namespace: str | None = None



from ...agent import (
    EngineEnvironment,
    EngineUri,
    Goal,
    Operation,
    OutputType,
    Role,
    Specification,
)
from ...agent.orchestrator import Orchestrator
from logging import Logger
from ...event.manager import EventManager
from ...memory.manager import MemoryManager
from ...model.entities import Input, TransformerEngineSettings
from ...model.manager import ModelManager
from ...tool.manager import ToolManager
from dataclasses import dataclass
from typing import Annotated, Optional, Union, get_args, get_origin


@dataclass(frozen=True, kw_only=True)
class Property:
    TYPE_MAP = {
        int: int.__name__,
        float: float.__name__,
        str: "string",
        bool: "boolean",
        list: "array",
        dict: "object",
        type(None): "null",
    }
    name: str
    data_type: str
    description: Optional[str] = None


class JsonSpecification(Specification):
    def __init__(
        self, output: Union[type, list[Property]], role: str, **kwargs
    ):
        if not isinstance(output, list):
            annotations = getattr(output, "__annotations__", None)
            assert annotations

            # Read annotated properties from output_class
            properties: list[Property] = []
            for name, type_info in annotations.items():
                if get_origin(type_info) is Annotated:
                    data_class, *metadata = get_args(type_info)
                    data_type = Property.TYPE_MAP.get(
                        data_class, data_class.__name__
                    )
                    description = metadata[0] if metadata else None
                    properties.append(
                        Property(
                            name=name,
                            data_type=data_type,
                            description=description.strip(),
                        )
                    )
        else:
            properties = output

        # Set specification defaults
        assert properties
        template_vars = {
            **(
                kwargs["template_vars"]
                if "template_vars" in kwargs and kwargs["template_vars"]
                else {}
            ),
            **{"output_properties": properties},
        }

        kwargs.setdefault("role", Role(persona=[role]))
        kwargs.setdefault("output_type", OutputType.JSON)
        kwargs.setdefault("template_vars", template_vars)
        super().__init__(**kwargs)


class JsonOrchestrator(Orchestrator):
    DEFAULT_FILE_NAME = "agent_json.md"

    def __init__(
        self,
        engine_uri: EngineUri,
        logger: Logger,
        model_manager: ModelManager,
        memory: MemoryManager,
        tool: ToolManager,
        event_manager: EventManager,
        output: Union[type, list[Property]],
        *,
        role: str,
        task: str,
        instructions: str,
        name: Optional[str] = None,
        rules: Optional[list[str]] = [],
        template_id: Optional[str] = None,
        settings: Optional[TransformerEngineSettings] = None,
        call_options: Optional[dict] = None,
        template_vars: Optional[dict] = None,
    ):
        specification = JsonSpecification(
            output=output,
            role=role,
            goal=Goal(task=task, instructions=[instructions]),
            rules=rules,
            template_id=template_id or JsonOrchestrator.DEFAULT_FILE_NAME,
            template_vars=template_vars,
        )
        super().__init__(
            logger,
            model_manager,
            memory,
            tool,
            event_manager,
            Operation(
                specification=specification,
                environment=EngineEnvironment(
                    engine_uri=engine_uri, settings=settings
                ),
            ),
            call_options=call_options,
        )

    async def __call__(self, input: Input, **kwargs) -> str:
        text_response = await super().__call__(input, **kwargs)
        return await text_response.to_json()



from abc import ABC, abstractmethod
from .....model import TextGenerationResponse, TextGenerationVendor
from .....model.entities import (
    GenerationSettings,
    Input,
    TransformerEngineSettings,
)
from .....model.nlp.text.generation import TextGenerationModel
from .....tool.manager import ToolManager
from dataclasses import replace
from logging import Logger
from tiktoken import encoding_for_model, get_encoding
from torch import Tensor
from transformers.tokenization_utils_base import BatchEncoding
from transformers import PreTrainedModel
from typing import Literal
from .....compat import override


class TextGenerationVendorModel(TextGenerationModel, ABC):
    _TIKTOKEN_DEFAULT_MODEL = "cl100k_base"

    def __init__(
        self,
        model_id: str,
        settings: TransformerEngineSettings | None = None,
        logger: Logger | None = None,
    ) -> None:
        settings = settings or TransformerEngineSettings()
        assert settings.access_token, "API key needed for vendor"
        settings = replace(settings, enable_eval=False)
        super().__init__(model_id, settings, logger)

    @abstractmethod
    def _load_model(self) -> PreTrainedModel | TextGenerationVendor:
        raise NotImplementedError()

    @property
    def supports_sample_generation(self) -> bool:
        return False

    @property
    def supports_token_streaming(self) -> bool:
        return True

    @property
    def uses_tokenizer(self) -> bool:
        return False

    def _tokenize_input(
        self,
        input: Input,
        context: str | None = None,
        tensor_format: Literal["pt"] = "pt",
        **kwargs,
    ) -> dict[str, Tensor] | BatchEncoding | Tensor:
        raise NotImplementedError()

    def input_token_count(
        self, input: Input, system_prompt: str | None = None
    ) -> int:
        try:
            encoding = encoding_for_model(self._model_id)
        except KeyError:
            encoding = get_encoding(self._TIKTOKEN_DEFAULT_MODEL)

        messages = self._messages(input, system_prompt, tool=None)

        total_tokens = 0
        for message in messages:
            total_tokens += len(encoding.encode(message.content or ""))
        return total_tokens

    @override
    async def __call__(
        self,
        input: Input,
        system_prompt: str | None = None,
        settings: GenerationSettings | None = None,
        *,
        tool: ToolManager | None = None,
    ) -> TextGenerationResponse:
        messages = self._messages(input, system_prompt, tool)
        streamer = await self._model(
            self._model_id,
            messages,
            settings,
            tool=tool,
            use_async_generator=settings.use_async_generator,
        )
        return TextGenerationResponse(
            streamer,
            settings=settings or GenerationSettings(),
            use_async_generator=settings.use_async_generator,
        )



from ..agent import (
    EngineEnvironment,
    EngineType,
    InputType,
    NoOperationAvailableException,
    Operation,
)
from ..agent.engine import EngineAgent
from ..agent.renderer import Renderer, TemplateEngineAgent
from ..event import Event, EventType
from ..event.manager import EventManager
from ..memory.manager import MemoryManager
from ..model import TextGenerationResponse
from ..model.engine import Engine
from ..model.entities import (
    EngineMessage,
    Input,
    Message,
    MessageRole,
)
from ..model.manager import ModelManager
from ..tool.manager import ToolManager
from contextlib import ExitStack
from dataclasses import asdict
from json import dumps
from logging import Logger
from typing import Any, Optional, Union, Type
from uuid import UUID, uuid4


class ObservableTextGenerationResponse(TextGenerationResponse):
    """Wrap a TextGenerationResponse to emit token and end events."""

    def __init__(
        self,
        response: TextGenerationResponse,
        event_manager: EventManager,
        model_id: str,
        tokenizer: Any | None,
    ) -> None:
        self._response = response
        self._event_manager = event_manager
        self._model_id = model_id
        self._tokenizer = tokenizer
        self._step = 0
        self._response.add_done_callback(self._on_consumed)

    async def _on_consumed(self) -> None:
        await self._event_manager.trigger(Event(type=EventType.STREAM_END))

    @property
    def input_token_count(self) -> int:
        return self._response.input_token_count

    def __aiter__(self) -> "ObservableTextGenerationResponse":
        self._response.__aiter__()
        return self

    async def __anext__(self) -> Any:
        token = await self._response.__anext__()
        token_str = token.token if hasattr(token, "token") else token
        token_id = getattr(token, "id", None)
        if token_id is None and self._tokenizer:
            ids = self._tokenizer.encode(token_str, add_special_tokens=False)
            token_id = ids[0] if ids else None

        await self._event_manager.trigger(
            Event(
                type=EventType.TOKEN_GENERATED,
                payload={
                    "token_id": token_id,
                    "model_id": self._model_id,
                    "token": token_str,
                    "step": self._step,
                },
            )
        )
        self._step += 1
        return token

    async def to_str(self) -> str:
        return await self._response.to_str()

    async def to_json(self) -> str:
        return await self._response.to_json()

    async def to(self, entity_class: type) -> Any:
        return await self._response.to(entity_class)


class Orchestrator:
    _id: UUID
    _name: Optional[str]
    _operations: list[Operation]
    _renderer: Renderer
    _total_operations: int
    _logger: Logger
    _model_manager: ModelManager
    _memory: MemoryManager
    _tool: ToolManager
    _event_manager: EventManager
    _engine_agents: dict[EngineEnvironment, EngineAgent] = {}
    _engines_stack: ExitStack = ExitStack()
    _operation_step: Optional[int] = None
    _model_ids: set[str] = set()
    _call_options: Optional[dict] = None
    _last_engine_agent: Optional[EngineAgent] = None
    _exit_memory: bool = True

    def __init__(
        self,
        logger: Logger,
        model_manager: ModelManager,
        memory: MemoryManager,
        tool: ToolManager,
        event_manager: EventManager,
        operations: Union[Operation, list[Operation]],
        *,
        call_options: Optional[dict] = None,
        exit_memory: bool = True,
        id: Optional[UUID] = None,
        name: Optional[str] = None,
        renderer: Optional[Renderer] = None,
    ):
        self._logger = logger
        self._model_manager = model_manager
        self._memory = memory
        self._tool = tool
        self._event_manager = event_manager
        self._operations = (
            [operations] if isinstance(operations, Operation) else operations
        )
        self._id = id or uuid4()
        self._exit_memory = exit_memory
        self._name = name
        self._renderer = renderer or Renderer()
        self._total_operations = len(self._operations)
        self._call_options = call_options

    @property
    def engine_agent(self) -> Optional[EngineAgent]:
        return self._last_engine_agent

    @property
    def engine(self) -> Optional[Engine]:
        return (
            self._last_engine_agent.engine if self._last_engine_agent else None
        )

    @property
    def id(self) -> UUID:
        return self._id

    @property
    def input_token_count(self) -> Optional[int]:
        return (
            self._last_engine_agent.input_token_count
            if self._last_engine_agent
            else None
        )

    @property
    def is_finished(self) -> bool:
        return (
            self._operation_step is not None
            and self._operation_step == self._total_operations - 1
        )

    @property
    def memory(self) -> MemoryManager:
        return self._memory

    @property
    def model_ids(self) -> set[str]:
        return self._model_ids

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def operations(self) -> list[Operation]:
        return self._operations

    @property
    def tool(self) -> ToolManager:
        return self._tool

    @property
    def event_manager(self) -> EventManager:
        return self._event_manager

    async def __call__(
        self, input: Input, **kwargs
    ) -> Union[TextGenerationResponse, str]:
        if self.is_finished:
            self._operation_step = 0

        # Pick next operation step
        operation_step = (
            self._operation_step + 1
            if self._operation_step
            and self._operation_step < self._total_operations
            else 0
            if not self._operation_step
            else None
        )
        self._operation_step = operation_step
        if self._operation_step is None:
            raise NoOperationAvailableException()

        # Load engine agent
        operation = self._operations[self._operation_step]
        environment_hash = dumps(asdict(operation.environment))
        assert self._engine_agents and environment_hash in self._engine_agents
        engine_agent = self._engine_agents[environment_hash]

        # Adapt tool manager
        if (
            engine_agent.engine.tokenizer
            and engine_agent.engine.tokenizer.eos_token
        ):
            self._tool.set_eos_token(engine_agent.engine.tokenizer.eos_token)

        await self._event_manager.trigger(
            Event(type=EventType.START, payload={"step": self._operation_step})
        )

        # Validate input
        input_type = operation.specification.input_type
        assert (
            input_type != InputType.TEXT
            or isinstance(input, str)
            or isinstance(input, Message)
            or isinstance(input, list)
        )

        if input_type == InputType.TEXT and isinstance(input, str):
            input = Message(role=MessageRole.USER, content=input)

        # Execute operation
        engine_args = {**(self._call_options or {}), **kwargs}
        result = await engine_agent(
            operation.specification, input, **engine_args
        )

        self._last_engine_agent = engine_agent

        # Inject tool results and continue ReACT inference if needed
        if not self._tool.is_empty:
            output = await result.to_str()

            await self._event_manager.trigger(
                Event(type=EventType.TOOL_PROCESS, payload={"output": output})
            )

            tool_calls, tool_results = self._tool(output)

            if tool_calls:
                for call in tool_calls:
                    await self._event_manager.trigger(
                        Event(
                            type=EventType.TOOL_EXECUTE, payload={"call": call}
                        )
                    )

            if tool_results:
                for res in tool_results:
                    await self._event_manager.trigger(
                        Event(
                            type=EventType.TOOL_RESULT, payload={"result": res}
                        )
                    )

            tool_messages = (
                [
                    Message(
                        role=MessageRole.TOOL,
                        name=r.name,
                        arguments=r.arguments,
                        content=r.result,
                    )
                    for r in tool_results
                ]
                if tool_results
                else None
            )

            self._logger.debug(f"Tool result messages: {tool_messages}")

            if tool_messages:
                result = await engine_agent(
                    operation.specification, tool_messages, **engine_args
                )

        await self._event_manager.trigger(Event(type=EventType.END))

        if isinstance(result, TextGenerationResponse):
            result = ObservableTextGenerationResponse(
                result,
                self._event_manager,
                engine_agent.engine.model_id,
                engine_agent.engine.tokenizer,
            )

        return result

    async def __aenter__(self):
        first_agent: Optional[TemplateEngineAgent] = None
        model_ids: list[str] = []
        for operation in self._operations:
            # Load engine with environment
            environment = operation.environment
            environment_hash = dumps(asdict(environment))
            if environment_hash not in self._engine_agents:
                model_ids.append(environment.engine_uri.model_id)
                engine = (
                    self._model_manager.load_engine(
                        environment.engine_uri, environment.settings
                    )
                    if environment.type == EngineType.TEXT_GENERATION
                    else None
                )
                if not engine:
                    raise NotImplementedError()

                self._engines_stack.enter_context(engine)
                agent = TemplateEngineAgent(
                    engine,
                    self._memory,
                    self._tool,
                    self._event_manager,
                    self._renderer,
                    name=self._name,
                    id=self._id,
                )
                self._engine_agents[environment_hash] = agent
                if not first_agent:
                    first_agent = agent

        self._last_engine_agent = first_agent
        self._model_ids = set(model_ids)
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[Any],
    ):
        if (
            self._last_engine_agent
            and self._last_engine_agent.output
            and (
                self._memory.has_permanent_message
                or self._memory.has_recent_message
            )
        ):
            previous_message = Message(
                role=MessageRole.ASSISTANT,
                content=await self._last_engine_agent.output.to_str(),
            )
            await self._memory.append_message(
                EngineMessage(
                    agent_id=self._id,
                    model_id=self._last_engine_agent.engine.model_id,
                    message=previous_message,
                )
            )

        if self._exit_memory:
            self._memory.__exit__(exc_type, exc_value, traceback)

        return self._engines_stack.__exit__(exc_type, exc_value, traceback)



from io import StringIO
from re import compile, escape, Pattern
from transformers import AutoTokenizer
from transformers.generation import StoppingCriteria


class KeywordStoppingCriteria(StoppingCriteria):
    _buffer: StringIO
    _tokenizer: AutoTokenizer
    _pattern: Pattern
    _keywords: list[str]
    _keyword_count: int

    def __init__(
        self,
        keywords: list[str],
        tokenizer: AutoTokenizer,
        all_must_be_present: bool = False,
    ):
        assert keywords
        escaped_keywords = [escape(k) for k in keywords]
        self._pattern = compile(
            r"^" + "".join(f"(?=.*{k})" for k in escaped_keywords) + r".*$"
            if all_must_be_present
            else r"(" + "|".join(escaped_keywords) + r")$"
        )
        self._buffer = StringIO()
        self._tokenizer = tokenizer
        self._keywords = keywords
        self._keyword_count = len(keywords)

    def __call__(self, input_ids, scores, **kwargs):
        token_id = input_ids[0][-1]
        token = self._tokenizer.decode(token_id, skip_special_tokens=False)
        self._buffer.write(token)
        buffer_contents = self._buffer.getvalue()
        return (
            self._keyword_count == 1
            and buffer_contents.endswith(self._keywords[0])
        ) or bool(self._pattern.search(buffer_contents))
